import uuid

from jupyter_ai_magics.providers import *
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import AIMessage
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.runnables import RunnableSerializable

from aita.agent.utils import get_model
from aita.datasource.base import DataSource
from aita.datasource.catalog import Catalog
from aita.prompt.base import BasePromptTemplate, BasicContextPromptTemplate


class AitaAgent:
    type: str = "aita"
    description: str = "Aita Agent is your best friend!"
    config: Dict[str, Any]
    model: RunnableSerializable
    model_parameters: Optional[Dict]
    prompt_template: Optional[ChatPromptTemplate]
    context_prompt: Optional[ChatPromptTemplate]
    context_prompt_template: List[ChatPromptTemplate]
    output_prompt: Optional[PromptTemplate]
    output_parser: Optional[PydanticOutputParser]

    def __init__(
        self,
        model: Union[str, RunnableSerializable],
        model_parameters: Optional[Dict] = None,
        **kwargs,
    ):
        # configs
        self.config = {
            "configurable": {
                "user_id": kwargs.get("user_id", str(uuid.uuid4())),
                "thread_id": kwargs.get("thread_id", str(uuid.uuid4())),
            }
        }

        # model
        self.model_parameters = model_parameters
        if isinstance(model, str):
            self.model: RunnableSerializable = get_model(model, model_parameters)
        else:
            self.model = model

        # prompt
        self.prompt_template = None
        self.base_prompt = self._set_base_prompt()
        self.context_prompt_template = []
        self.context_prompt = self._set_context_prompt()
        self.output_prompt = None
        self.output_parser = None

    def _create_chain(self, **kwargs) -> RunnableSerializable:
        if self.output_parser:
            return self.output_prompt | self.model | self.output_parser
        return self.prompt_template | self.model

    @staticmethod
    def _set_base_prompt(base_prompt: Optional[Union[str, ChatPromptTemplate]] = None):
        if not base_prompt:
            base_prompt = BasePromptTemplate
        if isinstance(base_prompt, str):
            base_prompt = ChatPromptTemplate.from_messages(
                [
                    ("system", base_prompt),
                ]
            )
        return base_prompt

    def set_base_prompt(self, base_prompt: Optional[Union[str, ChatPromptTemplate]] = None):
        self.base_prompt = self._set_base_prompt(base_prompt)
        return self

    @staticmethod
    def _set_context_prompt(context_prompt: Optional[Union[str, ChatPromptTemplate]] = None):
        if not context_prompt:
            context_prompt = BasicContextPromptTemplate
        if isinstance(context_prompt, str):
            context_prompt = ChatPromptTemplate.from_messages(
                [
                    ("system", context_prompt),
                ]
            )
        return context_prompt

    def set_context_prompt(self, context_prompt: Optional[Union[str, ChatPromptTemplate]] = None):
        self.context_prompt = self._set_context_prompt(context_prompt)
        return self

    def add_prompt_context(self, prompt_context: str):
        context_prompt = self.context_prompt.partial(prompt_context=prompt_context)
        self.context_prompt_template.append(context_prompt)

    def add_datasource(self, datasource: DataSource):
        self.add_prompt_context(str(datasource.get_metadata()))
        return self

    def add_data_catalog(self, catalog: Catalog):
        self.add_prompt_context(str(catalog))
        return self

    def _build_prompt(self, question: str):
        self.prompt_template = ChatPromptTemplate.from_messages(self.base_prompt.messages)
        self.prompt_template.extend(self.context_prompt_template)
        self.prompt_template.append(("user", question))

    def stream(self, question: Optional[str] = "", display: Optional[bool] = True):
        self._build_prompt(question)
        chain = self._create_chain()
        events = chain.stream(self.config)
        if display:
            for event in events:
                print(event.content, end="", flush=True)
        if self.output_parser:
            chain = self.output_prompt | self.model | self.output_parser
            events = self._parse_output(chain, events)
        return events

    def invoke(self, question: Optional[str] = "", display: Optional[bool] = True):
        self._build_prompt(question)
        chain = self._create_chain()
        results = chain.invoke(self.config)
        if display:
            for result in results:
                print(result.content, end="", flush=True)
        if self.output_parser:
            chain = self.output_prompt | self.model | self.output_parser
            results = self._parse_output(chain, results)
        return results

    def embed(self, query: str) -> list:
        assert hasattr(self.model, "embed"), "Model does not support embedding"
        return self.model.embed(query)

    def get_current_state(self) -> None:
        return None

    def _parse_output(self, chain, result: dict, max_iterations=10):
        iteration = 0
        while iteration < max_iterations:
            iteration += 1
            try:
                return chain.invoke({"messages": result["messages"]}, self.config)
            except OutputParserException as e:
                result["messages"] += [
                    AIMessage(
                        content=f"Error: {repr(e)}\n please fix your mistakes.",
                    )
                ]
        return result

    def set_output_parser(self, output_parser: BaseModel):
        self.output_parser = PydanticOutputParser(pydantic_object=output_parser)
        self.output_prompt = PromptTemplate(
            template="Return output in required format with given messages.\n{format_instructions}\n{messages}\n",
            input_variables=["messages"],
            partial_variables={"format_instructions": self.output_parser.get_format_instructions()},
        )
        return self



from jupyter_ai_magics.providers import *
from jupyter_ai_magics.utils import decompose_model_id, get_lm_providers
from langchain_core.runnables import RunnableSerializable


def get_model(model_id: str, model_parameters: Optional[Dict]) -> Optional[RunnableSerializable]:
    providers = get_lm_providers()
    provider_id, local_model_id = decompose_model_id(model_id, providers)
    if provider_id is None or provider_id not in providers:
        return None
    Provider = providers[provider_id]
    provider_params = {"model_id": local_model_id}
    model_parameters = model_parameters or {}
    return Provider(**provider_params, **model_parameters)



from aita.prompt.utils import get_sql_for_database


class BasicPrompt(object):
    def __init__(self, *args, **kwargs):
        # used to avoid empty init function in 0-shot prompt
        pass

    def format_target(self, example: dict):
        return self.format_question(example) + "\nSELECT "

    def format_question(self, example: dict):
        raise NotImplementedError()

    def get_extra_info(self, db_id):
        return None


class SQLPrompt(BasicPrompt):
    template_info = "/* Given the following database feature: */\n" "{}"
    template_question = "/* Answer the following: {} */"

    def format_question(self, example: dict):
        sqls = get_sql_for_database(example["path_db"])

        prompt_info = self.template_info.format("\n\n".join(sqls))
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n\n".join(prompt_components)
        return prompt


class TextPrompt(BasicPrompt):
    template_info = "Given the following database feature:\n" "{}"
    template_question = "Answer the following: {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}: {', '.join(_.feature)}" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class NumberSignPrompt(BasicPrompt):
    template_info = (
        "### Complete sqlite SQL query only and with no explanation\n"
        "### SQLite SQL tables, with their properties:\n"
        "#\n"
        "{}\n"
        "#"
    )
    template_question = "### {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"# {_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class BaselinePrompt(BasicPrompt):
    template_info = "{}\nForeign_keys={}\n"
    template_question = 'Q: "{}"'

    def format_question(self, example: dict):
        # schemas
        schemas = "\n".join(
            [f"Table {_.name}, columns = {_.feature}" for _ in example["tables"]]
        ).replace("'", "")
        # foreign_keys
        foreign_keys = list()
        for table in example["tables"]:
            for pair_str in table["table_info"]["foreign_key"]:
                a, b = [_.strip() for _ in pair_str[1:-1].split(",")]
                foreign_keys.append(f"{a}={b}")

        # format prompt
        prompt_info = self.template_info.format(schemas, str(foreign_keys).replace("'", ""))
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example) + "\nA: SELECT "


class InstructionPrompt(BasicPrompt):
    template_info = (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        '### Instruction:\nWrite a sql to answer the question "{}"\n\n### Input:\n{}\n'
    )
    template_question = "### Response:"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(example["question"], schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            # TODO: extra_info should be after info
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class TextWithForeignKeyPrompt(BasicPrompt):
    template_info = (
        "Given the following database feature:\n" "{} \n" "And their foreign keys:\n" "{}"
    )
    template_question = "Answer the following: {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}: {', '.join(_.feature)}" for _ in example["tables"]])
        # foreign_keys
        foreign_keys = list()
        for table in example["tables"]:
            for pair_str in table["table_info"]["foreign_key"]:
                a, b = [_.strip() for _ in pair_str[1:-1].split(",")]
                foreign_keys.append(f"{a}={b}")
        foreign_keys = f"{', '.join(foreign_keys)}"

        prompt_info = self.template_info.format(schemas, foreign_keys)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class NumberSignWithForeignKeyPrompt(BasicPrompt):
    template_info = (
        "### Complete sqlite SQL query only and with no explanation\n"
        "### SQLite SQL tables, with their properties:\n"
        "#\n"
        "{}\n"
        "#\n"
        "### Their foreign keys:\n"
        "#\n"
        "{}\n"
        "#"
    )
    template_question = "### {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"# {_.name}({', '.join(_.feature)})" for _ in example["tables"]])
        # foreign_keys
        foreign_keys = list()
        for table in example["tables"]:
            for pair_str in table["table_info"]["foreign_key"]:
                a, b = [_.strip() for _ in pair_str[1:-1].split(",")]
                foreign_keys.append(f"{a}={b}")
        foreign_keys = f"# Foreign_keys=({', '.join(foreign_keys)})"

        prompt_info = self.template_info.format(schemas, foreign_keys)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class BaselineWithoutForeignKeyPrompt(BasicPrompt):
    template_info = "{}\n"
    template_question = 'Q: "{}"'

    def format_question(self, example: dict):
        # schemas
        schemas = "\n".join(
            [f"Table {_.name}, columns = {_.feature}" for _ in example["tables"]]
        ).replace("'", "")

        # format prompt
        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example) + "\nA: SELECT "


class InstructionWithForeignKeyPrompt(BasicPrompt):
    template_info = (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        '### Instruction:\nWrite a sql to answer the question "{}"\n\n### Input:\n{}\nForeign Keys:{}\n'
    )
    template_question = "### Response:"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}({', '.join(_.feature)})" for _ in example["tables"]])
        # foreign_keys
        foreign_keys = list()
        for table in example["tables"]:
            for pair_str in table["table_info"]["foreign_key"]:
                a, b = [_.strip() for _ in pair_str[1:-1].split(",")]
                foreign_keys.append(f"{a}={b}")
        foreign_keys = f"{', '.join(foreign_keys)}"

        prompt_info = self.template_info.format(example["question"], schemas, foreign_keys)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            # TODO: extra_info should be after info
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class SQLWithRulePrompt(BasicPrompt):
    template_info = "/* Given the following database feature: */\n" "{}"
    template_question = "/* Answer the following with no explanation: {} */"

    def format_question(self, example: dict):
        sqls = get_sql_for_database(example["path_db"])

        prompt_info = self.template_info.format("\n\n".join(sqls))
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n\n".join(prompt_components)
        return prompt


class TextWithRulePrompt(BasicPrompt):
    template_info = "Given the following database feature:\n" "{}"
    template_question = "Answer the following with no explanation: {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}: {', '.join(_.feature)}" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class NumberSignWithoutRulePrompt(BasicPrompt):
    template_info = (
        "### Complete sqlite SQL query\n"
        "### SQLite SQL tables, with their properties:\n"
        "#\n"
        "{}\n"
        "#"
    )
    template_question = "### {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"# {_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class InstructionWithRulePrompt(BasicPrompt):
    template_info = (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        '### Instruction:\nWrite a sql only and with no explanation to answer the question "{}"\n\n### Input:\n{}\n'
    )
    template_question = "### Response:"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(example["question"], schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            # TODO: extra_info should be after info
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt


class SQLCOTPrompt(BasicPrompt):
    template_info = "/* Given the following database feature: */\n" "{}"
    template_question = "/* Let's think step by step. Answer the following: {} */"

    def format_question(self, example: dict):
        sqls = get_sql_for_database(example["path_db"])

        prompt_info = self.template_info.format("\n\n".join(sqls))
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n\n".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example)


class TextCOTPrompt(BasicPrompt):
    template_info = "Given the following database feature:\n" "{}"
    template_question = "Let's think step by step. Answer the following: {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}: {', '.join(_.feature)}" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example)


class NumberSignCOTPrompt(BasicPrompt):
    template_info = (
        "### Let's think step by step. Complete sqlite SQL query only and with no explanation\n"
        "### SQLite SQL tables, with their properties:\n"
        "#\n"
        "{}\n"
        "#"
    )
    template_question = "### {}"

    def format_question(self, example: dict):
        schemas = "\n".join([f"# {_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example)


class InstructionCOTPrompt(BasicPrompt):
    template_info = (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        '### Instruction:\nLet\'s think step by step. Write a sql to answer the question "{}"\n\n### Input:\n{}\n'
    )
    template_question = "### Response:"

    def format_question(self, example: dict):
        schemas = "\n".join([f"{_.name}({', '.join(_.feature)})" for _ in example["tables"]])

        prompt_info = self.template_info.format(example["question"], schemas)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            # TODO: extra_info should be after info
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt

    def format_target(self, example: dict):
        return self.format_question(example)


class CBRPrompt(BasicPrompt):
    template_info = (
        "# The following are the table names and column names needed to generate SQL:\n"
        "Tables: {}\n"
        "Columns: *, {}\n"
        "Foreign keys: {}"
    )
    template_question = '# translate "{}" into SQL query only and with no explanation:'

    def format_question(self, example: dict):
        tables = ", ".join([f"{_.name}" for _ in example["tables"]])
        columns = ", ".join([f"{_.name}.{col}" for _ in example["tables"] for col in _.feature])
        # foreign_keys
        foreign_keys = list()
        for table in example["tables"]:
            for pair_str in table["table_info"]["foreign_key"]:
                a, b = [_.strip() for _ in pair_str[1:-1].split(",")]
                foreign_keys.append(f"{a}={b}")
        foreign_keys = f"{', '.join(foreign_keys)}"

        prompt_info = self.template_info.format(tables, columns, foreign_keys)
        prompt_extra_info = self.get_extra_info(example["db_id"])
        prompt_question = self.template_question.format(example["question"])

        if prompt_extra_info is None or prompt_extra_info == "":
            prompt_components = [prompt_info, prompt_question]
        else:
            prompt_components = [prompt_info, prompt_extra_info, prompt_question]

        prompt = "\n".join(prompt_components)
        return prompt



"""Sidebar component for the app."""

import reflex as rx

from aita_lab import styles
from aita_lab.components.chat_navbar import navbar
from aita_lab.states.base import BaseState

# Icons for the sidebar
ICONS = {
    "Chat": "message-square-more",
    "Data Catalog": "library",
    "Data Source": "database",
    "Document": "file-code",
    "Prompt Template": "text",
    "Settings": "settings",
    "Tool": "hammer",
    "Agent": "bot",
    "Vector Store": "package",
}


def sidebar_header() -> rx.Component:
    """Sidebar header.

    Returns:
        The sidebar_chat_history header component.
    """
    return rx.hstack(
        # The logo.
        rx.color_mode_cond(
            rx.image(src="/aita_black.svg", height="3em"),
            rx.image(src="/aita_white.svg", height="3em"),
        ),
        rx.spacer(),
        rx.cond(
            BaseState.sidebar_displayed,
            rx.link(
                rx.button(
                    rx.icon("github"),
                    color_scheme="gray",
                    variant="soft",
                ),
                href="https://github.com/project-aita/aita",
                target="_blank",
            ),
        ),
        align="center",
        width="100%",
        border_bottom=styles.border,
        padding_x="1em",
        padding_y="2em",
    )


def sidebar_footer() -> rx.Component:
    """Sidebar footer.

    Returns:
        rx.Component: The sidebar_chat_history footer component.
    """
    return rx.chakra.hstack(
        rx.chakra.link(
            rx.chakra.center(
                rx.chakra.image(
                    src="/paneleft.svg",
                    height="2.5em",
                    padding="0.5em",
                ),
                bg="transparent",
                border_radius=styles.border_radius,
                **styles.hover_accent_bg,
            ),
            on_click=BaseState.toggle_sidebar_displayed,
            transform=rx.cond(~BaseState.sidebar_displayed, "rotate(180deg)", ""),
            transition="transform 0.5s, left 0.5s",
            position="relative",
            left=rx.cond(BaseState.sidebar_displayed, "13.5em", "1em"),
            **styles.overlapping_button_style,
        ),
        rx.chakra.spacer(),
        width="100%",
        border_top=styles.border,
        padding="1em",
    )


def sidebar_item(text: str, url: str) -> rx.Component:
    """Sidebar item.

    Args:
        text: The text of the item.
        url: The URL of the item.

    Returns:
        rx.Component: The sidebar_chat_history item component.
    """
    # Whether the item is active.
    active = (rx.State.router.page.path == url.lower()) | (
        (rx.State.router.page.path == "/") & text == "Chat"
    )

    return rx.link(
        rx.hstack(
            rx.icon(ICONS[text], color="gray", size=24),
            rx.text(
                text,
                opacity=rx.cond(BaseState.sidebar_displayed, 1, 0),
                visibility=rx.cond(BaseState.sidebar_displayed, "visible", "hidden"),
                position=rx.cond(BaseState.sidebar_displayed, "relative", "absolute"),
                transition="opacity 0.3s, visibility 0.3s",
            ),
            bg=rx.cond(
                active,
                rx.color("accent", 2),
                "transparent",
            ),
            border=rx.cond(
                active,
                f"1px solid {rx.color('accent', 6)}",
                f"1px solid {rx.color('gray', 6)}",
            ),
            color=rx.cond(
                active,
                styles.accent_text_color,
                styles.text_color,
            ),
            align="center",
            border_radius=styles.border_radius,
            width="100%",
            padding="1em",
        ),
        rx.cond(BaseState.sidebar_displayed, rx.cond(text == "Chat" and active, navbar())),
        href=url,
        width="100%",
    )


def sidebar() -> rx.Component:
    """The sidebar_chat_history.

    Returns:
        The sidebar_chat_history component.
    """
    # Get all the decorated pages and add them to the sidebar_chat_history.
    from reflex.page import get_decorated_pages

    return rx.box(
        rx.vstack(
            sidebar_header(),
            rx.vstack(
                *[
                    sidebar_item(
                        text=page.get("title", page["route"].strip("/").capitalize()),
                        url=page["route"],
                    )
                    for page in get_decorated_pages()
                ],
                width="100%",
                overflow_y="auto",
                align_items="flex-start",
                padding="1em",
            ),
            rx.spacer(),
            sidebar_footer(),
            height="100dvh",
        ),
        display=["none", "none", "block"],
        max_width=styles.sidebar_width,
        width=rx.cond(BaseState.sidebar_displayed, "100%", "100px"),
        transition="width 0.5s",
        height="100%",
        position="sticky",
        top="0px",
        border_right=styles.border,
    )



from typing import Optional

import importlib
import random

import reflex as rx
from langchain_core.runnables.graph import Edge, Node

from aita.agent.factory import AgentFactory, get_supported_agents
from aita.agent.graph import GraphAgent
from aita_lab.states.graph import Graph


def get_node_position(node: Node):
    if node.id == "__start__":
        return {"x": 20, "y": 20}
    elif node.id == "agent":
        return {"x": 20, "y": 100}
    elif node.id == "__end__":
        return {"x": 20, "y": 200}
    elif node.id == "tools":
        return {"x": 200, "y": 100}
    else:
        return {"x": random.randint(0, 500), "y": random.randint(0, 500)}


def create_agent_graph_node(node: Node):
    return {
        "id": node.id,
        "data": {"label": node.id},
        "position": get_node_position(node),
    }


def create_agent_graph_edge(id, edge: Edge):
    return {"id": id, "source": edge.source, "target": edge.target, "animated": True}


class Agent(rx.Model):
    id: Optional[str]
    name: str
    description: Optional[str]
    tools: Optional[list[str]]


class AgentState(rx.State):
    agents: list[Agent] = []
    is_open: bool = False
    current_agent: Optional[Agent] = None
    current_agent_graph: Optional[Graph] = None

    def open_agent(self, is_open: bool, agent: Agent):
        self.is_open = is_open
        passthrough_agent = AgentFactory.create_agent(agent["name"], model="")
        if isinstance(passthrough_agent, GraphAgent):
            graph = passthrough_agent.get_graph()
            self.current_agent_graph = Graph(
                nodes=[create_agent_graph_node(node) for _, node in graph.nodes.items()],
                edges=[create_agent_graph_edge(id, edge) for id, edge in enumerate(graph.edges)],
            )

    @rx.var
    def agent_names(self) -> list[str]:
        return [agent.name for agent in self.agents]

    def get_custom_agents(self):
        return []

    def on_load(self):
        self.agents = [
            Agent(name=agent.name, description=agent.value.description)
            for agent in get_supported_agents() + self.get_custom_agents()
        ]

    def create_agent(self, graph: Graph):
        state_graph = GraphAgent.init_state_graph()
        tools = []
        for node in graph.nodes:
            if node["type"] == "tool":
                tool_name = node["data"]["label"]
                tool_cls = importlib.import_module(f"aita.tool.{tool_name}")
                tools.append(tool_cls)

        state_graph.set_entry_point("agent")
        state_graph.add_edge("tools", "agent")

        agent = GraphAgent(type="custom", tools=tools, model=self.current_model, graph=state_graph)
        return



"""The home page of the app."""

import reflex as rx

from aita_lab import styles
from aita_lab.templates import template


@template(route="/document", title="Document")
def document() -> rx.Component:
    """The document page.

    Returns:
        The UI for the document page.
    """
    with open("README.md", encoding="utf-8") as readme:
        content = readme.read()
    return rx.markdown(content, component_map=styles.markdown_style)



from typing import Dict

from pandas import DataFrame

from aita.agent.graph import GraphAgent
from aita.tool.pandas_tool import PandasTool
from aita.tool.python_tool import PythonTool
from aita.tool.sql_tool import ConvertToPandasTool


class PandasAgent(GraphAgent):
    type: str = "pandas"
    description: str = (
        "A pandas agent that can use pandas tools to interact with pandas DataFrames."
    )

    def __init__(self, model: str, model_parameters: Dict = None):
        super().__init__(
            [
                PandasTool(),
                ConvertToPandasTool(),
            ],
            model,
            model_parameters,
        )

    def add_dataframe(self, dataframe: DataFrame):
        df_id = f"df_{id(dataframe)}"
        self.add_prompt_context(
            f"""
        dataframe name: {df_id}
        dataframe metadata: {dataframe.info}
        """
        )
        for tool in self.tools:
            if isinstance(tool, PythonTool):
                tool.update_script_context(script_context={df_id: dataframe})
        return self



import numpy as np

from aita.prompt.utils import count_tokens, get_tokenizer, jaccard_similarity


class BasicICLPrompt(object):
    NUM_EXAMPLE = None
    SEP_EXAMPLE = "\n\n"

    def __init__(self, tokenizer: str, *args, **kwargs):
        self.tokenizer = get_tokenizer(tokenizer)
        self.example_qualities = []
        self.pattern_similarities = []

    def record_example_quality(self, examples, target):
        quality_list = []
        for example in examples:
            quality_list.append(
                jaccard_similarity(example["query_skeleton"], target["query_skeleton"])
            )
        self.example_qualities.append(quality_list)

    def get_example_quality(self):
        if self.example_qualities:
            return np.mean([num for row in self.example_qualities for num in row])
        else:
            return 1

    def get_example_quality_for_each(self):
        if self.example_qualities:
            return [np.mean(row) for row in self.example_qualities]
        else:
            return []

    def record_pattern_similarity(self, examples, target):
        similarity_list = []
        for example in examples:
            similarity_list.append(
                jaccard_similarity(example["question_pattern"], target["question_pattern"])
            )
        self.pattern_similarities.append(similarity_list)

    def get_pattern_similarity(self):
        if self.pattern_similarities:
            return np.mean([num for row in self.pattern_similarities for num in row])
        else:
            return 1

    def format(
        self,
        target: dict,
        max_seq_len: int,
        max_ans_len: int,
        scope_factor: int,
        cross_domain=False,
        *args,
        **kwargs,
    ):
        # target question
        prompt_target = self.format_target(target)
        sum_tokens = count_tokens(prompt_target, tokenizer=self.tokenizer)

        if self.NUM_EXAMPLE != 0:
            # example questions
            examples = self.get_examples(
                target, self.NUM_EXAMPLE * scope_factor, cross_domain=cross_domain
            )
            prompt_example = list()
            question = target["question"]
            example_prefix = self.get_example_prefix()
            selected_examples = []
            for example in examples:
                example_question = example["question"]
                # assert example_question != question, f"Example is the same with target question: {question}!, \n{target}\n{example}"
                if cross_domain:
                    assert target["db_id"] != example["db_id"]

                example_format = self.format_example(example)

                # count tokens and drop the example if exceed max_len
                forward_tokens = count_tokens(
                    example_prefix
                    + self.SEP_EXAMPLE.join(prompt_example + [example_format, prompt_target]),
                    tokenizer=self.tokenizer,
                )

                if forward_tokens + max_ans_len <= max_seq_len:
                    # add an example
                    prompt_example.append(example_format)
                    # update tokens
                    sum_tokens = forward_tokens
                    # record the selected examples
                    selected_examples.append(example)

                    if len(prompt_example) >= self.NUM_EXAMPLE:
                        break

            self.record_example_quality(selected_examples, target)
            self.record_pattern_similarity(selected_examples, target)

            n_valid_example = len(prompt_example)
            if len(prompt_example) > 0:
                prompt = example_prefix + self.SEP_EXAMPLE.join(prompt_example + [prompt_target])
            else:
                prompt = self.SEP_EXAMPLE.join(prompt_example + [prompt_target])
        else:
            n_valid_example = 0
            prompt = prompt_target

        response_clean = " ".join(target["query"].split())[len("SELECT ") :]
        return {
            "prompt_tokens": sum_tokens,
            "prompt": prompt,
            "response": response_clean,
            "n_examples": n_valid_example,
            "db_id": target["db_id"],
        }



from typing import Any

import inspect

from langchain_core.pydantic_v1 import BaseModel


def get_model_classes(model: Any) -> list:
    return inspect.getmembers(model, inspect.isclass)


def get_model_fields(model: BaseModel, field_name: str) -> BaseModel:
    return model.__fields__[field_name].default


def get_model_fields_as_dict(model: BaseModel) -> dict:
    d = {}
    for field, value in model.__fields__.items():
        d[field] = {
            "name": field,
            "required": value.required,
            "description": value.field_info.description,
        }
    return d



"""The data catalog page """

import reflex as rx

from aita_lab.states.catalog import CatalogState, CatalogTable
from aita_lab.templates import template


def catalog_search():
    return rx.flex(
        rx.input(
            placeholder="Search for a table or column",
            width="100%",
        ),
        rx.button("Search"),
        spacing="3",
        justify="center",
        align="center",
        width="100%",
    )


def add_data_catalog():
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.button(
                rx.flex("Add Data Catalog", rx.icon(tag="plus", width=24, height=24), spacing="3"),
                size="4",
                radius="full",
                on_click=CatalogState.toggle_dialog,
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(
                "Connect and Add Data Catalog",
                size="1",
                font_family="Inter",
                padding_top="1em",
            ),
            rx.dialog.description(
                "Select the data source to add a data catalog",
                size="2",
                mb="4",
                padding_bottom="1em",
            ),
        ),
    )


def show_catalog(catalog: CatalogTable):
    return rx.table.row(
        rx.table.cell(catalog.catalog_name),
        rx.table.cell(catalog.schema),
        rx.table.cell(catalog.table),
        rx.table.cell(catalog.datasource_name),
        rx.table.cell(
            rx.button(
                "Delete",
                on_click=lambda: CatalogState.delete_catalog(catalog),
            )
        ),
    )


def catalog_grid():
    return rx.fragment(
        rx.vstack(
            rx.box(
                add_data_catalog(),
            ),
            rx.hstack(
                rx.select.root(
                    rx.select.trigger(
                        placeholder=f"Sort By:",
                    ),
                    rx.select.content(
                        rx.select.group(
                            rx.select.item("Catalog", value="catalog_name"),
                            rx.select.item("Schema", value="feature"),
                            rx.select.item("Table", value="table"),
                            rx.select.item("Data Source", value="datasource_name"),
                        )
                    ),
                    size="3",
                    value=CatalogState.sort_value,
                    on_change=lambda sort_value: CatalogState.sort_values(sort_value),
                    font_family="Inter",
                ),
                width="100%",
                padding_top="2em",
                padding_bottom="1em",
            ),
            rx.table.root(
                rx.table.header(
                    rx.table.row(
                        rx.table.column_header_cell("Catalog Name"),
                        rx.table.column_header_cell("Schema"),
                        rx.table.column_header_cell("Table"),
                        rx.table.column_header_cell("Data Source"),
                        rx.table.column_header_cell("Delete"),
                    ),
                ),
                rx.table.body(rx.foreach(CatalogState.catalogs, show_catalog)),
                # variant="surface",
                size="3",
                width="100%",
                justify="stretch",
            ),
            margin_top="20px",
        ),
    )


@template(route="/catalog", title="Data Catalog", on_load=CatalogState.on_load())
def catalog():
    return rx.vstack(
        rx.heading("Data Catalog", size="8"),
        rx.text("View your data catalog"),
        rx.box(
            catalog_search(),
            catalog_grid(),
            width="100%",
        ),
        width="100%",
    )



"""The Agent Page."""

import reflex as rx

from aita_lab.components.reactflow import background, controls, react_flow
from aita_lab.states.agent import Agent, AgentState
from aita_lab.states.graph import GraphState
from aita_lab.states.tool import ToolState
from aita_lab.templates import template


def agent_card(agent: Agent):
    """Create an agent card."""
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.chakra.card(
                rx.chakra.text(
                    agent.description,
                    no_of_lines=3,
                ),
                header=rx.chakra.heading(agent.name, size="md", padding="2px"),
                direction="column",
                overflow="auto",
                height="300px",
                width="50%",
                margin_right="20px",
                cursor="pointer",
                _hover={"background_color": rx.color("gray", 2)},
            )
        ),
        rx.dialog.content(
            rx.dialog.title(agent.name, size="6"),
            rx.dialog.description(agent.description),
            rx.cond(
                AgentState.current_agent_graph is not None,
                rx.vstack(
                    rx.heading("Agent Flow", size="4"),
                    react_flow(
                        background(),
                        controls(),
                        nodes_connectable=True,
                        on_connect=lambda e: GraphState.on_connect(e),
                        nodes=AgentState.current_agent_graph.nodes,
                        edges=AgentState.current_agent_graph.edges,
                        fit_view=True,
                    ),
                    margin_top="20px",
                    height="30em",
                    width="100%",
                ),
            ),
            rx.flex(
                rx.dialog.close(
                    rx.button("Close", size="2"),
                ),
                justify="end",
            ),
        ),
        on_open_change=lambda is_open: AgentState.open_agent(is_open, agent),
    )


def content_grid():
    """Create a content grid."""
    return rx.vstack(
        rx.chakra.text("Click on an agent to view its flow", width="100", padding_y="4"),
        rx.chakra.flex(rx.foreach(AgentState.agents, agent_card, width="100"), width="100%"),
    )


def create_agent_flow():
    return rx.vstack(
        rx.heading("Create your own agent", size="6"),
        rx.hstack(
            rx.select(
                ToolState.tool_names,
                placeholder="Select a tool to add",
                on_change=GraphState.add_tool_node,
            ),
            rx.button(
                "Create",
                variant="solid",
                on_click=lambda: AgentState.create_agent(GraphState.graph),
            ),
            rx.button(
                "Clear",
                variant="soft",
                color_scheme="gray",
                on_click=GraphState.clear_graph,
            ),
            width="100%",
        ),
        react_flow(
            background(),
            controls(),
            nodes_draggable=True,
            nodes_connectable=True,
            on_nodes_change=lambda e: GraphState.on_nodes_change(e),
            on_connect=lambda e: GraphState.on_connect(e),
            nodes=GraphState.graph.nodes,
            edges=GraphState.graph.edges,
            fit_view=True,
            border=f"1px solid {rx.color('gray', 6)}",
            border_radius="0.375rem",
        ),
        margin_top="20px",
        height="30em",
        width="100%",
    )


@template(route="/agent", title="Agent", on_load=[AgentState.on_load, ToolState.on_load])
def agent() -> rx.Component:
    """The tool page.

    Returns:
        The UI for the agent page.
    """
    return rx.vstack(
        rx.heading("Agent", size="8"),
        rx.box(
            content_grid(),
            margin_top="20px",
            width="100%",
        ),
        create_agent_flow(),
        # make the page full width
        width="100%",
    )



from typing import Any

from abc import abstractmethod

import pyarrow as pa

from aita.datasource.base import DataSource
from aita.datasource.catalog import Catalog


class SqlDataSource(DataSource):
    type: str = "sql"

    @abstractmethod
    def connect(self) -> Any:
        raise NotImplementedError

    def execute(self, query: str, params=None):
        with self.connect().cursor() as cursor:
            cursor.execute(query, *(params or ()))
            return cursor.fetchall()

    @abstractmethod
    def get_metadata(self, **kwargs) -> Catalog:
        raise NotImplementedError

    def to_arrow(self, query: str, params=None):
        with self.connect().cursor() as cursor:
            cursor.execute(query, *(params or ()))
            return cursor.fetch_arrow_table()

    def to_pandas(self, query: str, params=None):
        return self.to_arrow(query, params).to_pandas()

    def ingest(self, table_name: str, data: pa.Table, **kwargs):
        with self.connect().cursor() as cursor:
            res = cursor.adbc_ingest(table_name, data, **kwargs)
            cursor.connection.commit()
            return res



from enum import Enum

from IPython.display import Image, display
from jupyter_ai_magics.providers import *
from langchain.tools.render import render_text_description
from langchain_core.messages import HumanMessage, ToolCall, ToolMessage
from langchain_core.runnables import RunnableConfig, RunnableLambda, RunnableSerializable
from langchain_core.tools import BaseTool
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.pregel import StateSnapshot

from aita.agent.base import AitaAgent
from aita.datasource.base import DataSource
from aita.states import MessageState


class ToolMode(str, Enum):
    """The mode of the tool call."""

    DISALLOWED = "disallowed"
    CONTINUOUS = "continuous"
    ONCE = "once"


def handle_tool_error(state) -> dict:
    error = state.get("error")
    tool_calls = state["messages"][-1].tool_calls
    return {
        "messages": [
            ToolMessage(
                content=f"Error: {repr(error)}\n please fix your mistakes.",
                tool_call_id=tc["id"],
            )
            for tc in tool_calls
        ]
    }


class GraphAgent(AitaAgent):
    tools: List[BaseTool]
    graph: StateGraph
    compiled_graph: CompiledGraph

    def __init__(
        self,
        tools: List[BaseTool],
        model: Union[RunnableSerializable, str],
        model_parameters: Optional[Dict] = None,
        graph: Optional[StateGraph] = None,
        **kwargs,
    ):
        self.tools = tools

        super().__init__(model, model_parameters, **kwargs)

        # build the graph, this has to happen after the prompt is built
        self.memory = MemorySaver()
        self.compiled_graph = self._build_graph(graph)

    def _bind_tools(self):
        if hasattr(self.model, "bind_tools"):
            return self.model.bind_tools(self.tools)
        else:
            rendered_tools = render_text_description(self.tools)
            tool_prompt = f"""
            You are an assistant that has access to the following set of tools. Here are the names and descriptions for each tool:

            {rendered_tools}

            """
            self.base_prompt.append(("system", tool_prompt))
            return self.model

    def _build_graph(self, graph: StateGraph) -> CompiledGraph:
        if graph:
            return graph.compile(checkpointer=self.memory, interrupt_before=["tools"])
        workflow = StateGraph(MessageState)

        # Define the two nodes we will cycle between
        workflow.add_node("agent", self.call_model)
        workflow.add_node("tools", self.build_tool_node(self.tools))

        # We now add a conditional edge
        workflow.add_conditional_edges(
            "agent",
            tools_condition,
        )

        workflow.set_entry_point("agent")
        workflow.add_edge("tools", "agent")
        return workflow.compile(checkpointer=self.memory, interrupt_before=["tools"])

    @staticmethod
    def init_state_graph():
        return StateGraph(MessageState)

    def get_graph(self):
        return self.compiled_graph.get_graph(self.config)

    def get_current_state(self) -> Optional[StateSnapshot]:
        return self.compiled_graph.get_state(self.config)

    def get_current_state_messages(self):
        current_state = self.get_current_state()
        if current_state:
            return current_state.values.get("messages")
        return []

    def get_current_tool_calls(self) -> List[ToolCall]:
        current_state_messages = self.get_current_state().values.get("messages")
        if current_state_messages and current_state_messages[-1].tool_calls:
            return current_state_messages[-1].tool_calls
        return []

    def add_datasource(self, datasource: DataSource):
        super().add_datasource(datasource)
        for tool in self.tools:
            if hasattr(tool, "datasource"):
                tool.datasource = datasource
        return self

    def _build_prompt(self, question: str):
        current_state = self.get_current_state()
        if current_state.next and current_state.next[0] == "tools":
            # We are in the tool node, but the user has asked a new question
            # We need to deny the tool call and continue with the user's question
            tool_calls = self.get_current_tool_calls()
            return {
                "messages": [
                    ToolMessage(
                        tool_call_id=tool_calls[0]["id"],
                        content=f"Tool call denied by user. Reasoning: '{question}'. Continue assisting, accounting for the user's input.",
                    )
                ]
            }
        else:
            self.prompt_template = ChatPromptTemplate.from_messages(self.base_prompt.messages)
            self.prompt_template.extend(self.context_prompt_template)
            self.prompt_template.append(
                MessagesPlaceholder(variable_name="messages", optional=True)
            )
            return {"messages": [HumanMessage(content=question)]}

    def stream(
        self,
        question: Optional[str] = "",
        tool_mode: str = ToolMode.DISALLOWED,
        max_iterations: int = 10,
        display=True,
    ):
        if not question and tool_mode != ToolMode.DISALLOWED and self.get_current_tool_calls():
            messages = None
        else:
            messages = self._build_prompt(question)
        events = self.compiled_graph.stream(messages, config=self.config, stream_mode="values")
        if display:
            _printed = set()
            self._print_graph_events(events, _printed)

        if tool_mode == ToolMode.CONTINUOUS:
            current_state = self.get_current_state()
            iterations = 0
            while current_state.next and iterations < max_iterations:
                iterations += 1
                events = self.compiled_graph.stream(None, config=self.config)
                if display:
                    print(f"Iteration {iterations}")
                    self._print_graph_events(events, _printed)
                current_state = self.get_current_state()
        if self.output_parser:
            chain = self.output_prompt | self.model | self.output_parser
            events = self._parse_output(chain, events, max_iterations=max_iterations)
        return events

    def invoke(
        self,
        question: Optional[str] = "",
        tool_mode: str = ToolMode.DISALLOWED,
        max_iterations: int = 10,
        display=True,
    ):
        if not question and tool_mode != ToolMode.DISALLOWED and self.get_current_tool_calls():
            messages = None
        else:
            messages = self._build_prompt(question)
        result = self.compiled_graph.invoke(messages, config=self.config)
        if display:
            _printed = set()
            self._print_graph_events(result, _printed)

        if tool_mode == ToolMode.CONTINUOUS:
            print("Starting the iterative invocation process.")
            current_state = self.get_current_state()
            iterations = 0
            while current_state.next and iterations < max_iterations:
                iterations += 1
                result = self.compiled_graph.invoke(None, config=self.config)
                if display:
                    print(f"Iteration {iterations}")
                    self._print_graph_events(result, _printed)
                current_state = self.get_current_state()
        if self.output_parser:
            chain = self.output_prompt | self.model | self.output_parser
            result = self._parse_output(chain, result, max_iterations=max_iterations)
        return result

    @staticmethod
    def build_tool_node(tools):
        return ToolNode(tools).with_fallbacks(
            [RunnableLambda(handle_tool_error)], exception_key="error"
        )

    def call_tool(self, tool_name: str, tool_id: Optional[str] = None, **kwargs):
        if not tool_name:
            raise ValueError("Tool name is required.")
        curr_tool_calls = self.get_current_tool_calls()
        if tool_id:
            tool_call = next((tc for tc in curr_tool_calls if tc["id"] == tool_id), None)
        else:
            tool_call = next((tc for tc in curr_tool_calls if tc["name"] == tool_name), None)
        if not tool_call:
            raise ValueError(f"Tool call {tool_name} not found in current state.")
        tool = next((t for t in self.tools if t.name == tool_name), None)
        if not tool:
            raise ValueError(f"Tool {tool_name} not found in the tool sets.")
        if kwargs.get("args"):
            tool_call["args"].update(kwargs["args"])
        res = tool.invoke(tool_call["args"], self.config)
        return res

    def update_tool(self, tool_id: str, tool_args: dict):
        if not tool_id:
            raise ValueError("Tool id is required.")
        current_state_messages = self.get_current_state_messages()
        curr_tool_calls = self.get_current_tool_calls()
        tool_call_index = next(
            (index for (index, tc) in enumerate(curr_tool_calls) if tc["id"] == tool_id), None
        )
        if tool_call_index is None:
            raise ValueError(f"Tool call {tool_id} not found in current state.")
        current_state_messages[-1].tool_calls[tool_call_index]["args"] = tool_args
        new_state = {"messages": [current_state_messages[-1]]}
        self.compiled_graph.update_state(self.config, new_state)

    def cancel_tool(self, tool_id: str):
        pass

    def call_model(self, state: MessageState, config: RunnableConfig):
        chain = self.prompt_template | self._bind_tools()
        response = chain.invoke(state, self.config)
        return {"messages": [response]}

    def _print_graph_events(self, events, printed, max_length=1500):
        if isinstance(events, dict):
            events = [events]
        for event in events:
            messages = self._get_event_message(event)
            for message in messages:
                if message.id not in printed:
                    msg_repr = message.pretty_repr(html=True)
                    if len(msg_repr) > max_length:
                        msg_repr = msg_repr[:max_length] + " ... (truncated)"
                    print(msg_repr)
                    printed.add(message.id)

    def _get_event_message(self, event):
        if "tools" in event:
            return event["tools"]["messages"]
        if "agent" in event:
            return event["agent"]["messages"]
        if "messages" in event:
            return event["messages"]
        return event

    def display_graph(self):
        display(Image(self.compiled_graph.get_graph().draw_mermaid_png()))



from typing import Any, Dict, List

import reflex as rx


class ReactFlowLib(rx.Component):
    """A component that wraps a react flow lib."""

    library = "reactflow"

    def _get_custom_code(self) -> str:
        return """import 'reactflow/dist/style.css';
        """


class ReactFlow(ReactFlowLib):
    tag = "ReactFlow"

    nodes: rx.Var[List[Dict[str, Any]]]

    edges: rx.Var[List[Dict[str, Any]]]

    fit_view: rx.Var[bool]

    nodes_draggable: rx.Var[bool]

    nodes_connectable: rx.Var[bool]

    nodes_focusable: rx.Var[bool]

    on_nodes_change: rx.EventHandler[lambda e0: [e0]]

    on_edges_change: rx.EventHandler[lambda e0: [e0]]

    on_connect: rx.EventHandler[lambda e0: [e0]]

    def _get_custom_code(self) -> str:
        return """import 'reactflow/dist/style.css';
        """


class Background(ReactFlowLib):
    tag = "Background"

    color: rx.Var[str]

    gap: rx.Var[int]

    size: rx.Var[int]

    variant: rx.Var[str]


class Controls(ReactFlowLib):
    tag = "Controls"


class ApplyNodeChanges(ReactFlowLib):
    tag = "applyNodeChanges"


react_flow = ReactFlow.create
background = Background.create
controls = Controls.create
apply_node_changes = ApplyNodeChanges.create



from typing import Optional

import json
import logging
import os
from pathlib import Path

import pandas as pd
import reflex as rx
from feast import FeatureView, FileSource
from feast.data_format import ParquetFormat
from feast.data_source import PushSource
from feast.feature_store import FeatureStore
from feast.field import Field
from feast.repo_config import RepoConfig
from feast.repo_operations import _prepare_registry_and_repo, apply_total_with_repo_instance
from feast.types import Array, Float32
from sqlmodel import select

from aita_lab.states.base import BaseState


class VectorStore(rx.Model, table=True):
    project_name: str
    online_store_type: str
    offline_store_type: Optional[str]
    online_store_configs: Optional[str]
    offline_store_configs: Optional[str]


class FeastFeatureView(rx.Model):
    name: str
    feature: str
    entities: Optional[str] = ""
    source: Optional[str] = ""


class VectorStoreState(BaseState):
    projects: list[VectorStore] = []
    project_names: list[str] = []

    project_name: str
    online_store_type: str
    offline_store_type: Optional[str] = ""
    online_store_configs: str
    offline_store_configs: Optional[str] = ""

    feature_view_name: str
    feature_name: str
    feature_entities: Optional[str] = None
    data_source_type: Optional[str] = ""

    vector_feature_views: list[FeastFeatureView] = []

    store_dialog_open: bool = False
    feature_dialog_open: bool = False

    _fs: Optional[FeatureStore] = None

    files: list[str] = []

    def open_feature_dialog(self):
        self.feature_dialog_open = not self.feature_dialog_open

    def open_store_dialog(self):
        self.store_dialog_open = not self.store_dialog_open

    def set_project(self, project_name: str):
        self.project_name = project_name
        print(f"Project name: {self.project_name}")
        self._load_project()

    def _load_project(self):
        project = next(
            (project for project in self.projects if project.project_name == self.project_name),
            None,
        )
        if project:
            repo_config = self._get_feast_repo_config(
                project_name=project.project_name,
                online_store_type=project.online_store_type,
                online_store_configs=project.online_store_configs,
                offline_store_type=project.offline_store_type,
                offline_store_configs=project.offline_store_configs,
            )
            self._fs = FeatureStore(config=repo_config)
            self.vector_feature_views = self._get_feature_views()

    def _get_feast_repo_config(
        self,
        project_name,
        online_store_type,
        online_store_configs,
        offline_store_type: Optional[str] = None,
        offline_store_configs: Optional[str] = None,
    ):
        return RepoConfig(
            project=project_name,
            registry="data/vector.db",
            provider="local",
            online_config={
                **json.loads(online_store_configs),
                "type": online_store_type,
            },
            offline_config={
                **(json.loads(offline_store_configs) if offline_store_configs else {}),
                "type": offline_store_type,
            },
            entity_key_serialization_version=3,
        )

    def create_store(self):
        repo_config = self._get_feast_repo_config(
            project_name=self.project_name,
            online_store_type=self.online_store_type,
            online_store_configs=self.online_store_configs,
            offline_store_type=self.offline_store_type,
            offline_store_configs=self.offline_store_configs,
        )
        repo_path = Path(os.path.join(Path.cwd(), "data"))
        try:
            project, registry, repo, store = _prepare_registry_and_repo(repo_config, repo_path)
            apply_total_with_repo_instance(store, project, registry, repo, True)
            self._fs = FeatureStore(config=repo_config)

            # save the project to the database
            with rx.session() as session:
                session.add(
                    VectorStore(
                        project_name=self.project_name,
                        online_store_type=self.online_store_type,
                        offline_store_type=self.offline_store_type,
                        online_store_configs=self.online_store_configs,
                        offline_store_configs=self.offline_store_configs,
                    )
                )
                session.commit()
            logging.info("Feature store created")
        except Exception as e:
            logging.error(f"Error creating feature store: {e}")
            raise e
        finally:
            self.open_store_dialog()

    def _create_feature_schema(self):
        return [Field(name=self.feature_name, dtype=Array(Float32))]

    def _create_feature_source(self):
        return PushSource(
            name=self.feature_view_name,  # feature view name is used to reference the source
            batch_source=FileSource(
                file_format=ParquetFormat(),
                path="data/feature.parquet",
                timestamp_field="event_timestamp",
            ),
        )

    def create_vector_feature(self):
        self._fs.apply(
            FeatureView(
                name=self.feature_view_name,
                entities=self.feature_entities if self.feature_entities else [],
                schema=self._create_feature_schema(),
                source=self._create_feature_source(),
            )
        )
        logging.info(f"Feature View {self.feature_view_name} created")
        self.open_feature_dialog()

    def _get_feature_views(self) -> list[FeastFeatureView]:
        vector_feature_views = []
        for feature_view in self._fs.list_feature_views():
            feature_spec = self._fs.get_feature_view(feature_view.name)
            vector_feature_views.append(
                FeastFeatureView(
                    name=feature_spec.name,
                    entities=", ".join([entity.name for entity in feature_spec.entity_columns]),
                    feature=", ".join([feature.name for feature in feature_spec.features]),
                    source=feature_spec.stream_source.name,
                )
            )
        return vector_feature_views

    def load_store(self):
        with rx.session() as session:
            self.projects = session.exec(select(VectorStore)).all()
            self.project_names = [project.project_name for project in self.projects]

        # load the features from the first project
        if self.projects and len(self.projects) > 0:
            self.project_name = self.projects[0].project_name
            repo_config = self._get_feast_repo_config(
                project_name=self.project_name,
                online_store_type=self.projects[0].online_store_type,
                online_store_configs=self.projects[0].online_store_configs,
                offline_store_type=self.projects[0].offline_store_type,
                offline_store_configs=self.projects[0].offline_store_configs,
            )
            self._fs = FeatureStore(config=repo_config)
            self.vector_feature_views = self._get_feature_views()

        logging.info("Feature store loaded")

    def push_source_to_feature(self, feature_view: dict):
        (name, entities, feature, source) = (
            feature_view["name"],
            feature_view["entities"],
            feature_view["feature"],
            feature_view["source"],
        )
        root_dir = rx.get_upload_dir()
        source_dir = f"{root_dir}/{source}/"
        feature_df = pd.read_parquet(source_dir)
        if feature_df.empty:
            logging.error(f"Error loading feature: {source} is empty")
            return
        if not all(col in feature_df.columns for col in ["event_timestamp", feature, entities]):
            logging.error(
                f"Error loading feature: {source} is missing required columns: event_timestamp, {feature}, {entities}"
            )
            return

        try:
            self._fs.push(source, feature_df)
            logging.info(f"Source {source} pushed to feature.")
        except Exception as e:
            logging.error(f"Error loading feature: {e}")
            raise e

    def _retrieve_vector_features(self, feature: str, query, top_k: int = 3) -> dict:
        logging.info(f"Retrieving online documents for {feature} with query {query}")
        response = self._fs.retrieve_online_documents(feature=feature, query=query, top_k=top_k)
        return response.to_dict()

    async def handle_upload(self, files: list[rx.UploadFile]):
        """Handle the upload of file(s).

        Args:
            files: The uploaded files.
        """
        root_dir = rx.get_upload_dir()
        source_dir = f"{root_dir}/{self.feature_view_name}"
        if not os.path.exists(source_dir):
            os.makedirs(source_dir)

        for file in files:
            upload_data = await file.read()
            outfile = Path(f"{source_dir}/{file.filename}")

            # Save the file.
            with outfile.open("wb") as file_object:
                file_object.write(upload_data)

            # Update the files var.
            self.files.append(file.filename)

    def on_load(self) -> None:
        self.load_store()



"""Styles for the app."""

import reflex as rx

border_radius = "0.375rem"
border = f"1px solid {rx.color('gray', 6)}"
text_color = rx.color("gray", 11)
accent_text_color = rx.color("accent", 10)
accent_color = rx.color("accent", 1)
hover_accent_color = {"_hover": {"color": accent_text_color}}
hover_accent_bg = {"_hover": {"background_color": accent_color}}
content_width_vw = "90vw"
sidebar_width = "18em"

template_page_style = {
    "padding_top": "5em",
    "padding_x": ["auto", "2em"],
    "flex": "1",
    "overflow_x": "hidden",
}

template_content_style = {
    "border_radius": border_radius,
    "margin_bottom": "2em",
    "min_height": "90vh",
}

link_style = {
    "color": accent_text_color,
    "text_decoration": "none",
    **hover_accent_color,
}

overlapping_button_style = {
    "background_color": "white",
    "border_radius": border_radius,
}

markdown_style = {
    "h1": lambda text: rx.heading(text, size="5", margin_y="1em"),
    "h2": lambda text: rx.heading(text, size="3", margin_y="1em"),
    "h3": lambda text: rx.heading(text, size="1", margin_y="1em"),
    "p": lambda text: rx.text(text, color="black", margin_y="1em"),
    "code": lambda text: rx.code(text, color_scheme="gray"),
    "codeblock": lambda text, **props: rx.code_block(text, **props, margin_y="1em"),
    "a": lambda text, **props: rx.link(
        text,
        **props,
        color="blue",
        _hover={"color": "red"},
        font_weight="bold",
        text_decoration="underline",
        text_decoration_color=accent_text_color,
    ),
    "table": lambda el: rx.table.root(el, size="1", width="40em"),
    "thead": lambda el: rx.table.header(el, border_bottom=border),
    "tr": lambda text: rx.table.row(text, border_bottom=border),
}

# Common styles for questions and answers.
shadow = "rgba(0, 0, 0, 0.15) 0px 2px 8px"
chat_margin = "20%"
message_style = dict(
    border_radius="5px",
    box_shadow=shadow,
    display="inline-block",
    margin_y="0.5em",
    padding_left="1em",
    padding_right="1em",
    max_width="44em",
)



from aita.tool.pandas_tool import PandasTool
from aita.tool.pyarrow_tool import ArrowTool
from aita.tool.python_tool import PythonTool
from aita.tool.spark_tool import ConvertPandasToSparkTool, SparkTool
from aita.tool.sql_tool import ConvertToArrowTool, ConvertToPandasTool, SqlQueryTool



from typer import Typer
from subprocess import Popen, PIPE

cli = Typer()


@cli.command()
def start():
    process = Popen(["reflex", "run"], shell=False)
    process.communicate()


@cli.command()
def migrate():
    process = Popen(["reflex", "migrate"], shell=False)
    process.communicate()


def main():
    cli()


if __name__ == "__main__":
    main()



import reflex as rx

from aita_lab.states.chat import ChatState


def chat_history_item(chat: str) -> rx.Component:
    """A sidebar_chat_history chat item.

    Args:
        chat: The chat item.
    """
    return rx.drawer.close(
        rx.hstack(
            rx.button(
                chat, on_click=lambda: ChatState.set_chat(chat), width="80%", variant="surface"
            ),
            rx.button(
                rx.icon(
                    tag="trash",
                    on_click=ChatState.delete_chat,
                    stroke_width=1,
                ),
                width="20%",
                variant="surface",
                color_scheme="red",
            ),
            width="100%",
        )
    )


def sidebar_chat_history(trigger) -> rx.Component:
    """The sidebar_chat_history component."""
    return rx.drawer.root(
        rx.drawer.trigger(trigger),
        rx.drawer.overlay(),
        rx.drawer.portal(
            rx.drawer.content(
                rx.vstack(
                    rx.heading("Chat history", color=rx.color("mauve", 11)),
                    rx.divider(),
                    rx.foreach(ChatState.chat_titles, lambda chat: chat_history_item(chat)),
                    align_items="stretch",
                    width="100%",
                ),
                top="auto",
                right="auto",
                height="100%",
                width="20em",
                padding="2em",
                background_color=rx.color("mauve", 2),
                outline="none",
            )
        ),
        direction="left",
    )


def create_chat_modal(trigger) -> rx.Component:
    """A modal to create a new chat."""
    return rx.dialog.root(
        rx.dialog.trigger(trigger),
        rx.dialog.content(
            rx.hstack(
                rx.input(
                    placeholder="Create a new chat with title...",
                    on_blur=ChatState.set_new_chat_title,
                    width=["15em", "20em", "30em", "30em", "30em", "30em"],
                ),
                rx.dialog.close(
                    rx.button(
                        "Create chat",
                        on_click=ChatState.create_chat,
                    ),
                ),
                background_color=rx.color("mauve", 1),
                spacing="2",
                width="100%",
            ),
        ),
    )


def navbar():
    return rx.box(
        rx.hstack(
            rx.hstack(
                create_chat_modal(rx.button("+ New chat", variant="solid")),
                sidebar_chat_history(
                    rx.button(
                        rx.icon(
                            tag="messages-square",
                            color=rx.color("mauve", 10),
                        ),
                        rx.text("History"),
                        variant="soft",
                        color_scheme="gray",
                    )
                ),
                width="100%",
                justify="between",
            ),
            justify_content="space-between",
            align_items="center",
        ),
        backdrop_filter="auto",
        backdrop_blur="lg",
        padding="12px",
        border_bottom=f"1px solid {rx.color('mauve', 3)}",
        background_color=rx.color("mauve", 2),
        position="sticky",
        top="0",
        z_index="100",
        align_items="center",
    )



from typing import List

try:
    import boto3
except ImportError:
    boto3 = None

from aita.datasource.base import DataSource
from aita.datasource.catalog import Catalog


class DynamodbDataSource(DataSource):
    def __init__(self, region_name: str = None):
        self.region_name = region_name
        self.client = boto3.client("dynamodb", region_name=region_name)

    def get_metadata(self, table_name: str) -> List[Catalog]:
        response = self.client.describe_table(TableName=table_name)
        return response["Table"]



"""The settings page."""

import reflex as rx

from aita_lab.templates import ThemeState, template


@template(route="/settings", title="Settings")
def settings() -> rx.Component:
    """The settings page.

    Returns:
        The UI for the settings page.
    """
    return rx.vstack(
        rx.heading("Settings", size="8"),
        rx.hstack(
            rx.text("Dark mode: "),
            rx.color_mode.switch(),
        ),
        rx.hstack(
            rx.text("Primary color: "),
            rx.select(
                [
                    "tomato",
                    "red",
                    "ruby",
                    "crimson",
                    "pink",
                    "plum",
                    "purple",
                    "violet",
                    "iris",
                    "indigo",
                    "blue",
                    "cyan",
                    "teal",
                    "jade",
                    "green",
                    "grass",
                    "brown",
                    "orange",
                    "sky",
                    "mint",
                    "lime",
                    "yellow",
                    "amber",
                    "gold",
                    "bronze",
                    "gray",
                ],
                value=ThemeState.accent_color,
                on_change=ThemeState.set_accent_color,
            ),
        ),
        rx.hstack(
            rx.text("Secondary color: "),
            rx.select(
                [
                    "gray",
                    "mauve",
                    "slate",
                    "sage",
                    "olive",
                    "sand",
                ],
                value=ThemeState.gray_color,
                on_change=ThemeState.set_gray_color,
            ),
        ),
        rx.text(
            "You can edit this page in ",
            rx.code("{your_app}/pages/settings.py"),
            size="1",
        ),
    )



"""The Tool Page."""

import reflex as rx

from aita_lab.states.tool import Tool, ToolState
from aita_lab.templates import template


def tool_card(tool: Tool):
    """Create a tool card."""
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.chakra.card(
                rx.chakra.text(
                    tool.description,
                    no_of_lines=3,
                ),
                header=rx.chakra.heading(tool.name, size="md"),
                # adjust the size and make it scrollable
                direction="column",
                overflow="auto",
                height="300px",
                width="200px",
                margin="20px",
                cursor="pointer",
                _hover={"background_color": rx.color("gray", 2)},
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(tool.name, size="6"),
            rx.dialog.description(tool.description),
            rx.cond(
                tool.args is not None,
                rx.vstack(
                    rx.heading("Tool Arguments", size="4"),
                    rx.foreach(
                        tool.args,
                        lambda arg_pair: rx.flex(
                            rx.text(
                                arg_pair[0],
                                as_="div",
                                size="2",
                                mb="1",
                                weight="bold",
                            ),
                            rx.cond(
                                arg_pair[1].description is not None,
                                rx.text(
                                    arg_pair[1].description,
                                ),
                            ),
                            direction="column",
                            spacing="3",
                        ),
                    ),
                    margin_top="20px",
                    width="100%",
                ),
            ),
            rx.flex(
                rx.dialog.close(
                    rx.button("Close", size="2"),
                ),
                justify="end",
            ),
        ),
    )


def content_grid():
    """Create a content grid."""
    return rx.chakra.flex(
        rx.foreach(
            ToolState.tools,
            lambda tool: tool_card(tool),
        ),
        flex_wrap="wrap",
    )


@template(route="/tool", title="Tool", on_load=ToolState.on_load())
def tool() -> rx.Component:
    """The tool page.

    Returns:
        The UI for the tool page.
    """
    return rx.vstack(
        rx.chakra.heading("Tool"),
        rx.chakra.text("A suite of tools to help you with analyzing your data."),
        rx.box(
            content_grid(),
            margin_top="20px",
            width="100%",
        ),
        # make the page full width
        width="100%",
    )



from enum import Enum

# get embedding models
from jupyter_ai_magics.embedding_providers import (
    BedrockEmbeddingsProvider,
    GPT4AllEmbeddingsProvider,
    HfHubEmbeddingsProvider,
)
from jupyter_ai_magics.providers import (
    AI21Provider,
    BedrockChatProvider,
    BedrockProvider,
    GPT4AllProvider,
    HfHubProvider,
    QianfanProvider,
    SmEndpointProvider,
    TogetherAIProvider,
)


class OpenAIEmbeddingsProvider:
    id = "openai"
    name = "OpenAI"
    models = [
        "text-embedding-ada-002",
        "text-embedding-3-small",
        "text-embedding-3-large",
    ]


class MistralAIEmbeddingsProvider:
    id = "mistralai"
    name = "MistralAI"
    models = [
        "mistral-embed",
    ]


class CohereEmbeddingsProvider:
    id = "cohere"
    name = "Cohere"
    models = [
        "embed-english-v2.0",
        "embed-english-light-v2.0",
        "embed-multilingual-v2.0",
        "embed-english-v3.0",
        "embed-english-light-v3.0",
        "embed-multilingual-v3.0",
        "embed-multilingual-light-v3.0",
    ]


class OpenAIProvider:
    name = "OpenAI"
    id = "openai"
    models = ["babbage-002", "davinci-002", "gpt-3.5-turbo-instruct"]


class ChatOpenAIProvider:
    name = "OpenAIChat"
    id = "openai-chat"
    models = [
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-0125",
        "gpt-3.5-turbo-0301",  # Deprecated as of 2024-06-13
        "gpt-3.5-turbo-0613",  # Deprecated as of 2024-06-13
        "gpt-3.5-turbo-1106",
        "gpt-3.5-turbo-16k",
        "gpt-3.5-turbo-16k-0613",  # Deprecated as of 2024-06-13
        "gpt-4",
        "gpt-4-turbo-preview",
        "gpt-4-0613",
        "gpt-4-32k",
        "gpt-4-32k-0613",
        "gpt-4-0125-preview",
        "gpt-4-1106-preview",
    ]


class AzureChatOpenAIProvider:
    id = "azure-chat-openai"
    name = "Azure OpenAI"
    models = ["*"]


class ChatNVIDIAProvider:
    id = "nvidia-chat"
    name = "NVIDIA"
    models = [
        "playground_llama2_70b",
        "playground_nemotron_steerlm_8b",
        "playground_mistral_7b",
        "playground_nv_llama2_rlhf_70b",
        "playground_llama2_13b",
        "playground_steerlm_llama_70b",
        "playground_llama2_code_13b",
        "playground_yi_34b",
        "playground_mixtral_8x7b",
        "playground_neva_22b",
        "playground_llama2_code_34b",
    ]


class GeminiProvider:
    id = "gemini"
    name = "Gemini"
    models = [
        "gemini-1.0-pro",
        "gemini-1.0-pro-001",
        "gemini-1.0-pro-latest",
        "gemini-1.0-pro-vision-latest",
        "gemini-pro",
        "gemini-pro-vision",
    ]


class AnthropicProvider:
    id = "anthropic"
    name = "Anthropic"
    models = [
        "claude-v1",
        "claude-v1.0",
        "claude-v1.2",
        "claude-2",
        "claude-2.0",
        "claude-instant-v1",
        "claude-instant-v1.0",
        "claude-instant-v1.2",
    ]


class ChatAnthropicProvider:
    id = "anthropic-chat"
    name = "ChatAnthropic"
    models = [
        "claude-2.0",
        "claude-2.1",
        "claude-instant-1.2",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ]


class CohereProvider:
    id = "cohere"
    name = "Cohere"
    models = [
        "command",
        "command-nightly",
        "command-light",
        "command-light-nightly",
        "command-r-plus",
        "command-r",
    ]


class ChatModelProvider(Enum):
    openai_chat = ChatOpenAIProvider
    openai = OpenAIProvider
    ai21 = AI21Provider
    amazon_bedrock = BedrockProvider
    amazon_bedrock_chat = BedrockChatProvider
    anthropic = AnthropicProvider
    anthropic_chat = ChatAnthropicProvider
    azure_chat_openai = AzureChatOpenAIProvider
    cohere = CohereProvider
    gemini = GeminiProvider
    gpt4all = GPT4AllProvider
    huggingface_hub = HfHubProvider
    nvidia_chat = ChatNVIDIAProvider
    qianfan = QianfanProvider
    sagemaker_endpoint = SmEndpointProvider
    together_ai = TogetherAIProvider


class EmbeddingModelProvider(Enum):
    bedrock = BedrockEmbeddingsProvider
    cohere = CohereEmbeddingsProvider
    gpt4all = GPT4AllEmbeddingsProvider
    openai = OpenAIEmbeddingsProvider
    mistralai = MistralAIEmbeddingsProvider
    huggingface_hub = HfHubEmbeddingsProvider



from typing import Annotated
from typing_extensions import TypedDict

from langchain_core.messages import AIMessage, HumanMessage
from langgraph.graph.message import add_messages


class MessageState(TypedDict):
    messages: Annotated[list, add_messages]

    def human_message(self, content: str):
        return [message for message in self.messages if isinstance(message, HumanMessage)]

    def ai_message(self, content: str):
        return [message for message in self.messages if isinstance(message, AIMessage)]



from adbc_driver_manager.dbapi import Connection
from sqlalchemy.engine import URL

from aita.datasource.sql import SqlDataSource


class BigqueryDataSource(SqlDataSource):
    name: str = "BigQuery"

    def connect(self) -> Connection:
        pass

    def __init__(self, credentials_path: str, project: str, dataset: str, **kwargs):
        url = URL.create(
            "bigquery", credentials_path=credentials_path, project=project, dataset=dataset
        )
        super().__init__(str(url))



from typing import Optional

import adbc_driver_sqlite.dbapi
import pyarrow as pa
from adbc_driver_manager.dbapi import Connection
from pydantic import Field

from aita.datasource.catalog import Catalog
from aita.datasource.sql import SqlDataSource


class SqliteDataSource(SqlDataSource):
    connection_url: str = Field(..., description="Connection URL")

    def __init__(
        self,
        connection_url: Optional[str] = None,
    ):
        super().__init__(connection_url=connection_url)

    def connect(self) -> Connection:
        return adbc_driver_sqlite.dbapi.connect(self.connection_url)

    def get_metadata(self, **kwargs) -> Catalog:
        with self.connect() as conn:

            catalogs: pa.Table = conn.adbc_get_objects(
                catalog_filter=kwargs.get("database", conn.adbc_current_catalog),
                table_name_filter=kwargs.get("table", None),
            ).read_all()
            catalog = catalogs.to_pylist()[0]
            return Catalog(**catalog)



from typing import List, Optional

import json

import reflex as rx

from aita.prompt.prompt_builder import prompt_factory

f = open("aita/prompt/templates/formatted_data.json")
data = json.load(f)


class PromptTemplate(rx.Model):
    prompt_repr: str
    k_shot: int
    example_format: str
    selector_type: str
    prompt_template_name: str
    prompt_lines: List[str]


class PromptTemplateState(rx.State):
    question: str = data["question"]
    prompt_template_names: List[str] = []
    prompt_templates: List[PromptTemplate] = []

    @staticmethod
    def load_prompt_templates_from_data():
        prompt_template_names = []
        prompt_templates = []
        for template in data["templates"]:
            prompt_repr = template["args"]["prompt_repr"]
            k_shot = template["args"]["k_shot"]
            example_format = template["args"]["example_type"]
            selector_type = template["args"]["selector_type"]
            prompt_template_name = f"{prompt_repr}_{k_shot}-SHOT_{selector_type}_{example_format}"
            prompt_lines = template["formatted_question"]["prompt"].split("\n")
            prompt_template = PromptTemplate(
                prompt_repr=prompt_repr,
                k_shot=k_shot,
                example_format=example_format,
                selector_type=selector_type,
                prompt_template_name=prompt_template_name,
                prompt_lines=prompt_lines,
            )
            prompt_template_names.append(prompt_template_name)
            prompt_templates.append(prompt_template)
        return prompt_template_names, prompt_templates

    def load_prompt_templates(self):
        self.prompt_template_names, self.prompt_templates = self.load_prompt_templates_from_data()

    @staticmethod
    def get_prompt_template(prompt_template_name: str) -> Optional[PromptTemplate]:
        _, prompt_templates = PromptTemplateState.load_prompt_templates_from_data()
        return next(
            (pt for pt in prompt_templates if pt.prompt_template_name == prompt_template_name), None
        )

    @staticmethod
    def build_prompt(prompt_template: PromptTemplate, embedding_model: str, target):
        prompt = prompt_factory(
            repr_type=prompt_template.prompt_repr,
            k_shot=0,
            example_format=prompt_template.example_format,
            selector_type=prompt_template.selector_type,
        )(tokenizer=embedding_model)
        target_prompt = prompt.format(target, 2048, 200, 100, cross_domain=False)
        return target_prompt["prompt"]

    def on_load(self):
        self.load_prompt_templates()



import json

dataset1 = json.load(open("./dataset1.json"))
dataset2 = json.load(open("./dataset2.json"))

formatted_data = {
    "question": "What are the average and minimum price (in Euro) of all products?",
    "templates": [],
}


def convert(data):
    d = {
        "args": data["args"],
        "costs": data["costs"],
    }

    for question in data["questions"]:
        if (
            question["response"]
            == "avg(age) , min(age) , max(age) FROM singer WHERE country = 'France'"
        ):
            d["formatted_question"] = question

    return d


for data in [dataset1, dataset2]:
    formatted_data["templates"].append(convert(data))
print(formatted_data)
json.dump(formatted_data, open("./formatted_data.json", "w"), indent=4)



import findspark
import pandas as pd
from pyspark.sql import SparkSession

from aita.agent.graph import GraphAgent
from aita.tool.python_tool import PythonTool
from aita.tool.spark_tool import ConvertPandasToSparkTool, SparkTool


class SparkAgent(GraphAgent):
    type: str = "pyspark"
    description: str = "A PySpark agent that can use PySpark tools to run PySpark scripts."
    spark_session: SparkSession

    def __init__(self, spark_configs: dict[str, str], model: str, model_parameters=None):
        self.init_session(spark_configs)
        super().__init__(
            [
                SparkTool(),
                ConvertPandasToSparkTool(),
            ],
            model,
            model_parameters,
        )
        for tool in self.tools:
            if isinstance(tool, PythonTool):
                tool.update_script_context(script_context={"spark_session": self.spark_session})

    def init_session(self, spark_configs: dict[str, str]):
        self.spark_session = self.create_spark_session(spark_configs)
        self.spark_session.conf.set("spark.sql.execution.arrow.enabled", "true")

    @staticmethod
    def create_spark_session(spark_configs: dict[str, str]):
        assert "master" in spark_configs, "master is required in spark_configs"
        assert "app_name" in spark_configs, "app_name is required in spark_configs"
        findspark.init()
        return (
            SparkSession.builder.master(spark_configs.get("master"))
            .appName(spark_configs.get("app_name"))
            .getOrCreate()
        )

    def add_pandas_dataframe(self, dataframe: pd.DataFrame):
        df_id = f"df_{id(dataframe)}"
        self.add_prompt_context(
            f"""
        dataframe name: {df_id}
        dataframe metadata: {dataframe.info}
        """
        )
        for tool in self.tools:
            if isinstance(tool, PythonTool):
                tool.update_script_context(script_context={df_id: dataframe})
        return self



from typing import Union

from abc import abstractmethod

from langchain_core.pydantic_v1 import BaseModel, Field

from aita.datasource.catalog import Catalog, Schema, Table


class DataSource(BaseModel):
    type: str = Field(..., description="Type of the data source")

    @abstractmethod
    def get_metadata(self, **kwargs) -> Union[Catalog, Schema, Table]:
        pass



"""The Prompt template Page."""

import reflex as rx

from aita_lab.states.prompt_template import PromptTemplate, PromptTemplateState
from aita_lab.templates import template


def content_grid():
    """Create a content grid."""
    return rx.chakra.flex(
        rx.foreach(
            PromptTemplateState.prompt_templates,
            lambda pt: prompt_card(pt),
            # lambda pt: rx.chakra.card(
            #     rx.chakra.vstack(
            #         rx.foreach(pt.prompt_lines, lambda line: rx.chakra.text(line, padding="2px")),
            #         align_items="flex-start",
            #     ),
            #     header=rx.chakra.heading(pt.prompt_template_name, size="md"),
            #     # adjust the size and make it scrollable
            #     direction="column",
            #     overflow="auto",
            #     height="300px",
            #     width="50%",
            #     margin_right="20px",
            # ),
        )
    )


def prompt_card(pt: PromptTemplate):
    """Create a prompt card."""
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.chakra.card(
                rx.chakra.vstack(
                    rx.foreach(pt.prompt_lines, lambda line: rx.chakra.text(line, padding="2px")),
                    align_items="flex-start",
                ),
                header=rx.chakra.heading(pt.prompt_template_name, size="md"),
                # adjust the size and make it scrollable
                direction="column",
                overflow="auto",
                height="300px",
                width="50%",
                margin_right="20px",
                cursor="pointer",
                _hover={"background_color": rx.color("gray", 2)},
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(pt.prompt_template_name, size="6"),
            rx.chakra.flex(
                rx.chakra.badge(f"k_shot: {pt.k_shot}"),
                justify="start",
                direction="row",
                padding="4px",
            ),
            rx.chakra.vstack(
                rx.foreach(pt.prompt_lines, lambda line: rx.text(line)),
                align_items="flex-start",
                font_size="sm",
                margin_top="10px",
            ),
            rx.flex(
                rx.dialog.close(
                    rx.button("Close", size="2"),
                ),
                justify="end",
            ),
        ),
    )


def render_question():
    return rx.box(
        rx.chakra.text("A list of prompt templates for question:", size="md"),
        rx.chakra.badge(
            PromptTemplateState.question, margin_top="10px", font_size="md", padding="2px"
        ),
    )


@template(route="/prompt_template", title="Prompt Template", on_load=PromptTemplateState.on_load())
def prompt_template() -> rx.Component:
    """The prompt template page.

    Returns:
        The UI for the prompt template page.
    """
    return rx.vstack(
        rx.heading("Prompt Template", size="8"),
        render_question(),
        rx.box(
            content_grid(),
            margin_top="20px",
            width="100%",
        ),
        # make the page full width
        width="100%",
    )



from typing import Dict

import pyarrow as pa

from aita.agent.graph import GraphAgent
from aita.tool.pyarrow_tool import ArrowTool
from aita.tool.python_tool import PythonTool
from aita.tool.sql_tool import ConvertToArrowTool


class ArrowAgent(GraphAgent):
    type: str = "pyarrow"
    description: str = "An Arrow agent that can use Arrow tools to interact with Arrow Tables."

    def __init__(self, model: str, model_parameters: Dict = None):
        super().__init__([ConvertToArrowTool(), ArrowTool()], model, model_parameters)

    def add_table(self, table: pa.Table):
        table_id = f"table_{id(table)}"
        self.add_prompt_context(
            f"""
        pyarrow table name: {table_id}
        pyarrow table metadata: {table.schema}
        """
        )
        for tool in self.tools:
            if isinstance(tool, PythonTool):
                tool.update_script_context(script_context={table_id: table})
        return self



import random

import numpy as np

from aita.prompt.utils import jaccard_similarity


class BasicExampleSelector(object):
    def __init__(self, data, *args, **kwargs):
        self.data = data
        self.train_json = self.data.get_train_json()
        self.db_ids = [d["db_id"] for d in self.train_json]
        self.train_questions = self.data.get_train_questions()

    def get_examples(self, question, num_example, cross_domain=False):
        pass

    def domain_mask(self, candidates: list, db_id):
        cross_domain_candidates = [
            candidates[i] for i in range(len(self.db_ids)) if self.db_ids[i] != db_id
        ]
        return cross_domain_candidates

    def retrieve_index(self, indexes: list, db_id):
        cross_domain_indexes = [i for i in range(len(self.db_ids)) if self.db_ids[i] != db_id]
        retrieved_indexes = [cross_domain_indexes[i] for i in indexes]
        return retrieved_indexes


class RandomExampleSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)
        random.seed(0)

    def get_examples(self, target, num_example, cross_domain=False):
        train_json = self.train_json
        indexes = list(range(len(train_json)))
        if cross_domain:
            indexes = self.domain_mask(indexes, target["db_id"])
        selected_indexes = random.sample(indexes, num_example)
        if cross_domain:
            selected_indexes = self.retrieve_index(selected_indexes, target["db_id"])
        return [train_json[index] for index in selected_indexes]


class CosineSimilarExampleSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)

        self.SELECT_MODEL = "sentence-transformers/all-mpnet-base-v2"
        # self.SELECT_MODEL = "sentence-transformers/bert-base-nli-mean-tokens"

        from sentence_transformers import SentenceTransformer

        self.bert_model = SentenceTransformer(self.SELECT_MODEL, device="cpu")
        self.train_embeddings = self.bert_model.encode(self.train_questions)

    def get_examples(self, target, num_example, cross_domain=False):
        target_embedding = self.bert_model.encode([target["question"]])
        # target_embedding = self.bert_model.embed_text([target["question"]]).cpu().detach().numpy()

        # find the most similar question in train dataset
        from sklearn.metrics.pairwise import cosine_similarity

        similarities = np.squeeze(
            cosine_similarity(target_embedding, self.train_embeddings)
        ).tolist()
        pairs = [
            (similarity, index) for similarity, index in zip(similarities, range(len(similarities)))
        ]

        train_json = self.train_json
        pairs_sorted = sorted(pairs, key=lambda x: x[0], reverse=True)
        top_pairs = list()
        for s, index in pairs_sorted:
            similar_db_id = train_json[index]["db_id"]
            if cross_domain and similar_db_id == target["db_id"]:
                continue
            if train_json[index]["question"] == target["question"]:
                continue
            top_pairs.append((index, s))
            if len(top_pairs) >= num_example:
                break

        return [train_json[index] for (index, s) in top_pairs]


class EuclideanDistanceExampleSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)

        self.SELECT_MODEL = "sentence-transformers/all-mpnet-base-v2"

        from sentence_transformers import SentenceTransformer

        self.bert_model = SentenceTransformer(self.SELECT_MODEL, device="cpu")
        self.train_embeddings = self.bert_model.encode(self.train_questions)

    def get_examples(self, target, num_example, cross_domain=False):
        target_embedding = self.bert_model.encode([target["question"]])

        # find the most similar question in train dataset
        from sklearn.metrics.pairwise import euclidean_distances

        distances = np.squeeze(
            euclidean_distances(target_embedding, self.train_embeddings)
        ).tolist()
        pairs = [(distance, index) for distance, index in zip(distances, range(len(distances)))]

        train_json = self.train_json
        pairs_sorted = sorted(pairs, key=lambda x: x[0])
        top_pairs = list()
        for d, index in pairs_sorted:
            similar_db_id = train_json[index]["db_id"]
            if cross_domain and similar_db_id == target["db_id"]:
                continue
            top_pairs.append((index, d))
            if len(top_pairs) >= num_example:
                break

        return [train_json[index] for (index, d) in top_pairs]


class EuclideanDistanceThresholdExampleSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)

        self.SELECT_MODEL = "sentence-transformers/all-mpnet-base-v2"
        # self.top_distances = list()
        self.threshold = 0.85

        from sentence_transformers import SentenceTransformer

        self.bert_model = SentenceTransformer(self.SELECT_MODEL, device="cpu")
        self.train_embeddings = self.bert_model.encode(self.train_questions)

    def get_examples(self, target, num_example, cross_domain=False):
        target_embedding = self.bert_model.encode([target["question"]])

        # find the most similar question in train dataset
        from sklearn.metrics.pairwise import euclidean_distances

        distances = np.squeeze(
            euclidean_distances(target_embedding, self.train_embeddings)
        ).tolist()
        pairs = [(distance, index) for distance, index in zip(distances, range(len(distances)))]

        train_json = self.train_json
        pairs_sorted = sorted(pairs, key=lambda x: x[0])
        top_pairs = list()
        for d, index in pairs_sorted:
            similar_db_id = train_json[index]["db_id"]
            if (cross_domain and similar_db_id == target["db_id"]) or d > self.threshold:
                continue
            top_pairs.append((index, d))
            # self.top_distances.append(d)
            if len(top_pairs) >= num_example:
                break
        # print("mean", np.mean(self.top_distances))    # 0.822
        # print("std", np.std(self.top_distances, ddof=1))  # 0.144
        # print("max", max(self.top_distances)) # 1.166

        return [train_json[index] for (index, d) in top_pairs]


class EuclideanDistancePreSkeletonSimilarThresholdSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)

        self.SELECT_MODEL = "sentence-transformers/all-mpnet-base-v2"
        self.threshold = 0.85

        from sentence_transformers import SentenceTransformer

        self.bert_model = SentenceTransformer(self.SELECT_MODEL, device="cpu")
        self.train_embeddings = self.bert_model.encode(self.train_questions)

    def get_examples(self, target, num_example, cross_domain=False):
        target_embedding = self.bert_model.encode([target["question"]])

        # find the most similar question in train dataset
        from sklearn.metrics.pairwise import euclidean_distances

        distances = np.squeeze(
            euclidean_distances(target_embedding, self.train_embeddings)
        ).tolist()
        pairs = [(distance, index) for distance, index in zip(distances, range(len(distances)))]

        train_json = self.train_json
        pairs_sorted = sorted(pairs, key=lambda x: x[0])
        top_pairs = list()
        for d, index in pairs_sorted:
            similar_db_id = train_json[index]["db_id"]
            if cross_domain and similar_db_id == target["db_id"]:
                continue
            # Skeleton similarity
            if (
                jaccard_similarity(train_json[index]["pre_skeleton"], target["pre_skeleton"])
                < self.threshold
            ):
                continue
            top_pairs.append((index, d))
            if len(top_pairs) >= num_example:
                break

        if len(top_pairs) < num_example:
            for d, index in pairs_sorted:
                similar_db_id = train_json[index]["db_id"]
                if cross_domain and similar_db_id == target["db_id"]:
                    continue
                # Skeleton similarity
                if (
                    jaccard_similarity(train_json[index]["pre_skeleton"], target["pre_skeleton"])
                    >= self.threshold
                ):
                    continue
                top_pairs.append((index, d))
                if len(top_pairs) >= num_example:
                    break

        return [train_json[index] for (index, d) in top_pairs]


class EuclideanDistancePreSkeletonSimilarPlusSelector(BasicExampleSelector):
    def __init__(self, data, *args, **kwargs):
        super().__init__(data)

        self.SELECT_MODEL = "sentence-transformers/all-mpnet-base-v2"

        from sentence_transformers import SentenceTransformer

        self.bert_model = SentenceTransformer(self.SELECT_MODEL, device="cpu")
        self.train_embeddings = self.bert_model.encode(self.train_questions)

    def get_examples(self, target, num_example, cross_domain=False):
        target_embedding = self.bert_model.encode([target["question"]])

        # find the most similar question in train dataset
        from sklearn.metrics.pairwise import euclidean_distances

        distances = np.squeeze(
            euclidean_distances(target_embedding, self.train_embeddings)
        ).tolist()
        train_json = self.train_json
        for i in range(len(train_json)):
            distances[i] -= jaccard_similarity(
                train_json[i]["pre_skeleton"], target["pre_skeleton"]
            )
        pairs = [(distance, index) for distance, index in zip(distances, range(len(distances)))]
        pairs_sorted = sorted(pairs, key=lambda x: x[0])
        top_pairs = list()
        for d, index in pairs_sorted:
            similar_db_id = train_json[index]["db_id"]
            if cross_domain and similar_db_id == target["db_id"]:
                continue
            top_pairs.append((index, d))
            if len(top_pairs) >= num_example:
                break

        return [train_json[index] for (index, d) in top_pairs]



from typing import Any, Dict, Sequence, Type, Union

import logging

from IPython import get_ipython
from IPython.core.interactiveshell import ExecutionResult, InteractiveShell
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_core.tools import BaseTool

log = logging.getLogger(__name__)


class PythonInput(BaseModel):
    script: str = Field(description="python script")


class PythonTool(BaseTool):
    """Tool for running python script in an IPython environment."""

    name: str = "run_ipython_script_tool"
    description: str = """
    Execute a python script in an IPython environment and return the result of the last expression.
    If the script is not correct, an error message will be returned.
    """
    args_schema: Type[BaseModel] = PythonInput

    ipython: InteractiveShell = None

    def __init__(self, /, **data: Any):
        super().__init__(**data)
        self.ipython = get_ipython()
        if not self.ipython:
            self.ipython = InteractiveShell()

    def _run(
        self,
        script,
    ) -> Union[str, Sequence[Dict[str, Any]], ExecutionResult]:
        """Execute the script, return the result or an error message."""
        try:
            result = self.ipython.run_cell(script)
            return result
        except Exception as e:
            return str(e)

    def update_script_context(self, script_context: Any):
        try:
            self.ipython.user_ns.update(script_context)
        except Exception as e:
            return str(e)



from typing import Any

import reflex as rx

from aita_lab import styles


def upload_render(files: list[str], handle_upload: Any):
    """The main view."""
    return rx.vstack(
        rx.upload(
            rx.vstack(
                rx.button(
                    "Select File",
                ),
                rx.text("Drag and drop files here or click to select files"),
                align="center",
            ),
            id="upload2",
            multiple=True,
            accept={
                "application/pdf": [".pdf"],
                "image/png": [".png"],
                "image/jpeg": [".jpg", ".jpeg"],
                "image/gif": [".gif"],
                "image/webp": [".webp"],
                "text/html": [".html", ".htm"],
            },
            max_files=5,
            disabled=False,
            on_drop=handle_upload(rx.upload_files(upload_id="upload2")),
            border=styles.border,
            padding="5em",
        ),
        rx.grid(
            rx.foreach(
                files,
                lambda file: rx.vstack(
                    rx.text(file),
                ),
            ),
            columns="2",
            spacing="1",
        ),
        padding="5em",
    )



"""Common templates used between pages in the app."""

from __future__ import annotations

from typing import Callable

import reflex as rx

from aita_lab import styles
from aita_lab.components.sidebar import sidebar

# Meta tags for the app.
default_meta = [
    {
        "name": "viewport",
        "content": "width=device-width, shrink-to-fit=no, initial-scale=1",
    },
]


def menu_item_link(text, href):
    return rx.menu.item(
        rx.link(
            text,
            href=href,
            width="100%",
            color="inherit",
        ),
        _hover={
            "color": styles.accent_color,
            "background_color": styles.accent_text_color,
        },
    )


def menu_button() -> rx.Component:
    """The menu button on the top right of the page.

    Returns:
        The menu button component.
    """
    from reflex.page import get_decorated_pages

    return rx.box(
        rx.menu.root(
            rx.menu.trigger(
                rx.button(
                    rx.icon("menu"),
                    variant="soft",
                )
            ),
            rx.menu.content(
                *[menu_item_link(page["title"], page["route"]) for page in get_decorated_pages()],
                rx.menu.separator(),
                menu_item_link("About", "https://github.com/project-aita/aita"),
                menu_item_link("Contact", "mailto:founders@oroject-aita.com"),
            ),
        ),
        position="fixed",
        right="2em",
        top="2em",
        z_index="500",
    )


class ThemeState(rx.State):
    """The state for the theme of the app."""

    accent_color: str = "iris"

    gray_color: str = "gray"


def template(
    route: str | None = None,
    title: str | None = None,
    description: str | None = None,
    meta: str | None = None,
    script_tags: list[rx.Component] | None = None,
    on_load: rx.event.EventHandler | list[rx.event.EventHandler] | None = None,
) -> Callable[[Callable[[], rx.Component]], rx.Component]:
    """The template for each page of the app.

    Args:
        route: The route to reach the page.
        title: The title of the page.
        description: The description of the page.
        meta: Additionnal meta to add to the page.
        on_load: The event handler(s) called when the page load.
        script_tags: Scripts to attach to the page.

    Returns:
        The template with the page content.
    """

    def decorator(page_content: Callable[[], rx.Component]) -> rx.Component:
        """The template for each page of the app.

        Args:
            page_content: The content of the page.

        Returns:
            The template with the page content.
        """
        # Get the meta tags for the page.
        all_meta = [*default_meta, *(meta or [])]

        def templated_page():
            return rx.hstack(
                sidebar(),
                rx.box(
                    rx.vstack(
                        page_content(),
                        rx.spacer(),
                        **styles.template_content_style,
                    ),
                    **styles.template_page_style,
                ),
                menu_button(),
                align="start",
                background=f"radial-gradient(circle at top right, {rx.color('accent', 2)}, {rx.color('mauve', 1)});",
                position="relative",
            )

        @rx.page(
            route=route,
            title=title,
            description=description,
            meta=all_meta,
            script_tags=script_tags,
            on_load=on_load,
        )
        def theme_wrap():
            return rx.theme(
                templated_page(),
                has_background=True,
                accent_color=ThemeState.accent_color,
                gray_color=ThemeState.gray_color,
            )

        return theme_wrap

    return decorator



"""Base state for the app."""

from typing import Optional

import reflex as rx


class User(rx.Model, table=True):
    """User Model"""

    name: str
    created_at: Optional[str]
    updated_at: Optional[str]


class BaseState(rx.State):
    """State for the app."""

    user: User = User(name="User")

    sidebar_displayed: bool = True

    @rx.var
    def origin_url(self) -> str:
        """Get the url of the current page.

        Returns:
            str: The url of the current page.
        """
        return self.router_data.get("asPath", "")

    def toggle_sidebar_displayed(self) -> None:
        """Toggle the sidebar_chat_history displayed."""
        self.sidebar_displayed = not self.sidebar_displayed

    def load_user(self) -> None:
        """Load the user."""
        with rx.session() as session:
            self.user = session.exec(rx.select(User).where(User.name == "User")).first()

    def on_load(self) -> None:
        """Load the state."""
        self.load_user()



import reflex as rx


class CodeEditor(rx.Component):
    library = "@uiw/react-codemirror"

    lib_dependencies: list[str] = ["@uiw/codemirror-extensions-langs"]

    tag = "CodeEditor"

    is_default = True

    value: rx.Var[str]

    height: rx.Var[str]

    extension: rx.Var[dict]
    on_change: rx.EventHandler[lambda value: [value]]


code_editor = CodeEditor.create



"""The chat page."""

import reflex as rx

from aita_lab import styles
from aita_lab.components.codeeditor import code_editor
from aita_lab.components.loading_icon import loading_icon
from aita_lab.components.model_selector import select_chat_model, select_embedding_model
from aita_lab.states.agent import AgentState
from aita_lab.states.chat import QA, ChatState
from aita_lab.states.datasource import DataSourceState
from aita_lab.states.prompt_template import PromptTemplateState
from aita_lab.states.vector_store import VectorStoreState
from aita_lab.styles import markdown_style, message_style
from aita_lab.templates import template


def message(qa: QA) -> rx.Component:
    """A single question/answer message.

    Args:
        qa: The question/answer pair.

    Returns:
        A component displaying the question/answer pair.
    """
    return rx.box(
        rx.box(
            rx.markdown(
                qa.question,
                background_color=rx.color("mauve", 4),
                color=rx.color("mauve", 12),
                component_map=markdown_style,
                **message_style,
            ),
            text_align="right",
            margin_top="1em",
        ),
        rx.box(
            rx.markdown(
                qa.answer,
                background_color=rx.color("accent", 4),
                color=rx.color("accent", 12),
                component_map=markdown_style,
                **message_style,
            ),
            text_align="left",
            padding_top="1em",
            width="100%",
        ),
        width="100%",
    )


def chat_history() -> rx.Component:
    """List all the messages in a single conversation."""
    return rx.vstack(
        rx.box(rx.foreach(ChatState.chats[ChatState.current_chat], message), width="100%"),
        py="8",
        flex="1",
        width="100%",
        max_width="50em",
        padding_x="4px",
        align_self="center",
        overflow_y="auto",
        overflow_x="hidden",
        padding_bottom="5em",
    )


def action_bar() -> rx.Component:
    """The action bar to send a new message."""
    return rx.center(
        rx.vstack(
            rx.chakra.form(
                rx.chakra.form_control(
                    rx.flex(
                        rx.radix.text_area(
                            placeholder="Type something...",
                            id="question",
                            enter_key_submit=True,
                            flex_grow=1,
                        ),
                        rx.button(
                            rx.cond(
                                ChatState.processing,
                                loading_icon(height="1em", width="1em"),
                                rx.text("Send"),
                            ),
                            type="submit",
                        ),
                        spacing="4",
                        align_items="center",
                        padding_x="16px",
                    ),
                    is_disabled=ChatState.processing,
                    width="100%",
                ),
                on_submit=ChatState.process_question,
                reset_on_submit=True,
                width="100%",
            ),
            rx.text(
                "Aita may return factually incorrect or misleading responses. Use discretion.",
                text_align="center",
                font_size=".75em",
                color=rx.color("mauve", 10),
            ),
            align_items="center",
            width="100%",
        ),
        position="sticky",
        bottom="0",
        left="0",
        padding_y="16px",
        backdrop_filter="auto",
        backdrop_blur="lg",
        border_top=f"1px solid {rx.color('mauve', 3)}",
        background_color=rx.color("mauve", 2),
        align_items="stretch",
        width="100%",
    )


def datasource_selector() -> rx.Component:
    """The datasource selector."""
    return rx.chakra.form(
        rx.text(
            "Data Source",
            asi_="div",
            mb="1",
            size="2",
            weight="bold",
        ),
        rx.chakra.form_control(
            rx.select(
                items=DataSourceState.datasource_names,
                value=ChatState.current_datasource,
                placeholder="Select a datasource",
                on_change=ChatState.set_current_datasource,
                width="12em",
            ),
            label="Datasource",
            width="100%",
        ),
    )


def prompt_template_selector() -> rx.Component:
    """The prompt template selector."""
    return rx.chakra.form(
        rx.chakra.form_control(
            rx.text(
                "Prompt Template",
                asi_="div",
                mb="1",
                size="2",
                weight="bold",
            ),
            rx.select(
                PromptTemplateState.prompt_template_names,
                value=ChatState.current_prompt_template.prompt_template_name,
                on_change=ChatState.set_current_prompt_template,
                width="14em",
                placeholder="Select a prompt template",
            ),
            rx.dialog.root(
                rx.dialog.content(
                    rx.flex(
                        select_embedding_model(
                            ChatState.current_embedding_model, ChatState.set_current_embedding_model
                        ),
                        rx.text(
                            "K-Shot",
                            asi_="div",
                            mb="1",
                            size="2",
                            weight="bold",
                        ),
                        rx.input(
                            placeholder="Enter the number of examples",
                            width="100%",
                            on_blur=ChatState.set_current_k_shot,
                            padding_top="2",
                        ),
                        rx.text(
                            "Vector Feature *",
                            asi_="div",
                            mb="1",
                            size="2",
                            weight="bold",
                        ),
                        rx.select.root(
                            rx.select.trigger(
                                placeholder="Select your feature",
                            ),
                            rx.select.content(
                                rx.select.group(
                                    rx.select.label("Select your feature"),
                                    rx.foreach(
                                        VectorStoreState.vector_feature_views,
                                        lambda x: rx.select.item(
                                            x.name, value=f"{x.name}:{x.feature}"
                                        ),
                                    ),
                                ),
                            ),
                            value=ChatState.current_vector_feature,
                            on_change=ChatState.set_current_vector_feature,
                        ),
                        rx.dialog.close(
                            rx.button(
                                "Confirm",
                                on_click=ChatState.close_vector_feature_dialog,
                            )
                        ),
                        spacing="2",
                        direction="column",
                    ),
                    padding="3",
                    margin_top="3",
                    border=styles.border,
                    border_radius=styles.border_radius,
                ),
                open=ChatState.vector_feature_dialog_open,
            ),
            label="Prompt Template",
            width="100%",
        ),
    )


def agent_selector() -> rx.Component:
    return rx.chakra.form(
        rx.chakra.form_control(
            rx.text(
                "Agent Type",
                asi_="div",
                mb="1",
                size="2",
                weight="bold",
            ),
            rx.select(
                AgentState.agent_names,
                value=ChatState.current_chat_agent_type,
                on_change=ChatState.set_current_chat_agent_type,
                width="14em",
                placeholder="Select an agent type",
            ),
            label="Agent Type",
            width="100%",
        )
    )


def tool_args() -> rx.Component:
    return rx.flex(
        rx.cond(
            ChatState.current_tool_id is not None and ChatState.current_tool_args.length() > 0,
            rx.foreach(
                ChatState.current_tool_args,
                lambda arg: rx.box(
                    rx.text(
                        arg.name,
                        asi_="div",
                        mb="1",
                        size="2",
                        weight="bold",
                    ),
                    code_editor(
                        value=arg.value,
                        on_change=lambda x: ChatState.update_current_tool_arg(arg.name, x),
                        width="100%",
                        min_height="20e",
                        language="python",
                        theme="material",
                        font_size="1em",
                        padding="4px",
                    ),
                    padding="6px",
                ),
            ),
        ),
        border_radius=styles.border_radius,
        direction="column",
    )


def tool_panel() -> rx.Component:
    return rx.flex(
        rx.flex(
            rx.select.root(
                rx.select.trigger(),
                rx.select.content(
                    rx.select.group(
                        rx.select.label("Select a tool"),
                        rx.foreach(
                            ChatState.current_tools_as_list,
                            lambda x: rx.select.item(x.id, value=x.id),
                        ),
                    ),
                ),
                value=ChatState.current_tool_id,
                on_change=ChatState.set_current_tool_by_id,
                padding="10px",
            ),
            rx.select.root(
                rx.select.trigger(),
                rx.select.content(
                    rx.select.group(
                        rx.select.label("Tool Name"),
                        rx.foreach(
                            ChatState.current_tools_as_list,
                            lambda x: rx.select.item(x.name, value=x.name),
                        ),
                    ),
                ),
                value=ChatState.current_tool_name,
                on_change=ChatState.set_current_tool_by_name,
                padding="10px",
            ),
            rx.chakra.button(
                rx.icon(tag="play"),
                size="xs",
                on_click=ChatState.run_tool,
            ),
            rx.chakra.button(
                rx.icon(tag="circle_stop"),
                size="xs",
                on_click=ChatState.cancel_tool,
            ),
            align="center",
            spacing="2",
            width="100%",
            padding="8px",
        ),
        rx.divider(),
        tool_args(),
        spacing="1",
        direction="column",
        background_color=rx.color("mauve", 3),
    )


def tool_kernel() -> rx.Component:
    """The code editor wrapper for running tools."""
    return rx.vstack(
        rx.chakra.heading(
            "Tool Kernel",
            size="md",
        ),
        tool_panel(),
        tool_output(),
        width="40em",
        height="85vh",
        padding="10px",
        gap="2",
        align_items="stretch",
        background_color=rx.color("mauve", 2),
        color=rx.color("mauve", 12),
        border=f"1px solid {rx.color('accent', 10)}",
        border_radius=styles.border_radius,
    )


def tool_output() -> rx.Component:
    return rx.box(
        rx.cond(
            ChatState.run_tool_output.show,
            rx.box(
                rx.badge(
                    "Output",
                    asi_="div",
                    mb="1",
                    size="3",
                    weight="bold",
                    padding="4px",
                ),
                rx.cond(
                    ChatState.run_tool_output.show,
                    rx.data_table(
                        data=ChatState.run_tool_output.data,
                        width="38em",
                        pagination=True,
                        search=True,
                        sort=True,
                        resizable=True,
                    ),
                ),
                width="100%",
                height="30em"
            ),
        ),
    )


@template(
    route="/",
    title="Chat",
    on_load=[
        ChatState.on_load,
        DataSourceState.on_load,
        PromptTemplateState.on_load,
        AgentState.on_load,
        VectorStoreState.on_load,
    ],
)
def chat() -> rx.Component:
    """The main app."""
    return rx.chakra.flex(
        rx.chakra.box(
            rx.chakra.hstack(
                select_chat_model(
                    ChatState.current_chat_model,
                    ChatState.set_current_chat_model,
                    style={"width": "12em"},
                ),
                prompt_template_selector(),
                datasource_selector(),
                agent_selector(),
                padding="4",
                width="100%",
                background_color=rx.color("mauve", 2),
                color=rx.color("mauve", 12),
                border_left=f"1px solid {rx.color('mauve', 3)}",
                top="0",
                spacing="4",
                align_items="stretch",
            ),
            rx.chakra.vstack(
                chat_history(),
                action_bar(),
                background_color=rx.color("mauve", 1),
                color=rx.color("mauve", 12),
                height="80vh",
                align_items="stretch",
                spacing="0",
                padding="20px",
                overflow_y="scroll",
            ),
            flex_grow=1,
        ),
        rx.chakra.grid_item(
            tool_kernel(),
            width="100%",
        ),
        h="100vh",
        width="100%",
        gap="2",
        direction="row",
        overflow_y="hidden",
    )



from typing import Type

from langchain_core.pydantic_v1 import BaseModel, Field

from aita.tool.python_tool import PythonTool


class PandasInput(BaseModel):
    script: str = Field(description="pandas script")


class PandasTool(PythonTool):
    """Tool for running Pandas analysis."""

    name: str = "pandas_tool"
    description: str = """
    Run a python script by using the Pandas library.
    If the script is not correct, an error message will be returned.

    Pandas dataframes are stored in the script context.
    """

    args_schema: Type[BaseModel] = PandasInput



class ReprType:
    CODE_REPRESENTATION = "SQL"
    TEXT_REPRESENTATION = "TEXT"
    OPENAI_DEMOSTRATION = "NUMBERSIGN"
    BASIC = "BASELINE"
    ALPACA_SFT = "INSTRUCTION"
    OPENAI_DEMOSTRATION_WFK = "NUMBERSIGNWFK"
    BASIC_WOFK = "BASELINEWOFK"
    TEXT_REPRESENTATION_WFK = "TEXTWFK"
    ALPACA_SFT_WFK = "INSTRUCTIONWFK"
    OPENAI_DEMOSTRATION_WORULE = "NUMBERSIGNWORULE"
    CODE_REPRESENTATION_WRULE = "SQLWRULE"
    ALPACA_SFT_WRULE = "INSTRUCTIONWRULE"
    TEXT_REPRESENTATION_WRULE = "TEXTWRULE"
    CODE_REPRESENTATION_COT = "SQLCOT"
    TEXT_REPRESENTATION_COT = "TEXTCOT"
    OPENAI_DEMOSTRATION_COT = "NUMBERSIGNCOT"
    ALPACA_SFT_COT = "INSTRUCTIONCOT"
    CBR = "CBR"


class ExampleType:
    ONLY_SQL = "ONLYSQL"
    QA = "QA"
    COMPLETE = "COMPLETE"
    QAWRULE = "QAWRULE"
    OPENAI_DEMOSTRATION_QA = "NUMBERSIGNQA"
    BASIC_QA = "BASELINEQA"


class SelectorType:
    COS_SIMILAR = "COSSIMILAR"
    RANDOM = "RANDOM"
    EUC_DISTANCE = "EUCDISTANCE"
    EUC_DISTANCE_THRESHOLD = "EUCDISTANCETHRESHOLD"
    EUC_DISTANCE_SKELETON_SIMILARITY_THRESHOLD = "EUCDISSKLSIMTHR"
    EUC_DISTANCE_QUESTION_MASK = "EUCDISQUESTIONMASK"
    EUC_DISTANCE_PRE_SKELETON_SIMILARITY_THRESHOLD = "EUCDISPRESKLSIMTHR"
    EUC_DISTANCE_PRE_SKELETON_SIMILARITY_PLUS = "EUCDISPRESKLSIMPLUS"
    EUC_DISTANCE_MASK_PRE_SKELETON_SIMILARITY_THRESHOLD = "EUCDISMASKPRESKLSIMTHR"
    EUC_DISTANCE_MASK_PRE_SKELETON_SIMILARITY_THRESHOLD_SHIFT = "EUCDISMASKPRESKLSIMTHRSHIFT"
    # TODO: from the same database



from enum import Enum

from aita.datasource.base import DataSource
from aita.datasource.bigquery import BigqueryDataSource
from aita.datasource.file import FileDataSource
from aita.datasource.mysql import MySqlDataSource
from aita.datasource.nosql import DynamodbDataSource
from aita.datasource.postgresql import PostgreSqlDataSource
from aita.datasource.snowflake import SnowflakeDataSource
from aita.datasource.sqlite import SqliteDataSource


class DataSourceProvider(Enum):
    mysql = MySqlDataSource
    postgresql = PostgreSqlDataSource
    bigquery = BigqueryDataSource
    snowflake = SnowflakeDataSource
    file = FileDataSource
    dynamodb = DynamodbDataSource
    sqlite = SqliteDataSource


def get_supported_datasources():
    return list(DataSourceProvider)


class DataSourceFactory:
    @staticmethod
    def create_datasource(datasource: str, *args, **kwargs) -> DataSource:
        if not hasattr(DataSourceProvider, datasource):
            raise ValueError(f"Unsupported datasource: {datasource}")

        datasource_class = DataSourceProvider[datasource].value
        return datasource_class(*args, **kwargs)



from enum import Enum

from aita.agent.arrow_agent import ArrowAgent
from aita.agent.base import AitaAgent
from aita.agent.pandas_agent import PandasAgent
from aita.agent.python_agent import PythonAgent
from aita.agent.spark_agent import SparkAgent
from aita.agent.sql import SqlAgent


class AgentProvider(Enum):
    aita = AitaAgent
    sql = SqlAgent
    pandas = PandasAgent
    pyarrow = ArrowAgent
    pyspark = SparkAgent
    python = PythonAgent
    embedding = AitaAgent


def get_supported_agents():
    return list(AgentProvider)


class AgentFactory:

    @staticmethod
    def create_agent(agent_type: str, *args, **kwargs) -> AitaAgent:
        if not agent_type or not hasattr(AgentProvider, agent_type):
            agent_class = AitaAgent
        else:
            agent_class = AgentProvider[agent_type].value
        return agent_class(*args, **kwargs)



from typing import Any, Dict, List, Optional

import json
import random
from functools import reduce

import reflex as rx
from reflex.state import MutableProxy, serialize_mutable_proxy

initial_nodes = [
    {
        "id": "1",
        "type": "input",
        "data": {
            "label": "__start__",
        },
        "position": {"x": 200, "y": 25},
    },
    {
        "id": "2",
        "data": {"label": "agent"},
        "position": {"x": 200, "y": 125},
    },
    {
        "id": "3",
        "type": "output",
        "data": {"label": "__end__"},
        "position": {"x": 200, "y": 250},
    },
]

initial_edges = [
    {"id": "e1-2", "source": "1", "target": "2", "label": "*", "animated": True},
    {"id": "e2-3", "source": "2", "target": "3", "label": "*", "animated": True},
]


def to_dict_if_proxy(obj):
    """Converts MutableProxy to dict if necessary"""
    if isinstance(obj, MutableProxy):
        json_obj = serialize_mutable_proxy(obj)
        return json.loads(json_obj)
    return obj


def handle_parent_expand(res: List[Any], update_item: Dict[str, Any]) -> None:
    for item in res:
        if item.get("id") == update_item.get("parentNode"):
            item["expanded"] = True
            break


def apply_changes(
    changes: List[Dict[str, Any]], elements: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    # we need this hack to handle the setNodes and setEdges function of the useReactFlow hook for controlled flows
    if any(c["type"] == "reset" for c in changes):
        return [c["item"] for c in changes if c["type"] == "reset"]

    init_elements = [c["item"] for c in changes if c["type"] == "add"]
    return reduce(lambda res, item: apply_change(res, item, changes), elements, init_elements)


def apply_change(
    res: List[Dict[str, Any]], item: Dict[str, Any], changes: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    current_changes = [c for c in changes if c["id"] == item["id"]]

    if not current_changes:
        res.append(item)
        return res

    update_item = item.copy()

    for current_change in current_changes:
        if current_change:
            if current_change["type"] == "select":
                update_item["selected"] = current_change["selected"]
            elif current_change["type"] == "position":
                if "position" in current_change:
                    update_item["position"] = current_change["position"]
                if "positionAbsolute" in current_change:
                    update_item["positionAbsolute"] = current_change["positionAbsolute"]
                if "dragging" in current_change:
                    update_item["dragging"] = current_change["dragging"]
                if update_item.get("expandParent"):
                    handle_parent_expand(res, update_item)
            elif current_change["type"] == "dimensions":
                if "dimensions" in current_change:
                    update_item["width"] = current_change["dimensions"]["width"]
                    update_item["height"] = current_change["dimensions"]["height"]
                if "updateStyle" in current_change:
                    update_item["style"] = {
                        **(update_item.get("style", {})),
                        **current_change["dimensions"],
                    }
                if "resizing" in current_change:
                    update_item["resizing"] = current_change["resizing"]
                if update_item.get("expandParent"):
                    handle_parent_expand(res, update_item)
            elif current_change["type"] == "remove":
                return res

    res.append(update_item)
    return res


class Graph(rx.Model):
    nodes: List[Dict[str, Any]]
    edges: List[Dict[str, Any]]


class GraphState(rx.State):
    """The app state."""

    graph: Graph = Graph(nodes=initial_nodes, edges=initial_edges)

    current_tool: Optional[str]

    def add_tool_node(self, tool):
        self.current_tool = tool
        new_node_id = f"{len(self.graph.nodes) + 1}"
        node_type = "tool"
        x = random.randint(0, 500)
        y = random.randint(0, 500)

        new_node = {
            "id": new_node_id,
            "type": node_type,
            "data": {"label": self.current_tool},
            "position": {"x": x, "y": y},
            "draggable": True,
        }
        self.graph.nodes.append(new_node)

    def _set_graph(self, graph: Graph):
        self.graph = graph

    def add_random_node(self):
        new_node_id = f"{len(self.graph.nodes) + 1}"
        node_type = random.choice(["default"])
        # Label is random number
        label = new_node_id
        x = random.randint(0, 500)
        y = random.randint(0, 500)

        new_node = {
            "id": new_node_id,
            "type": node_type,
            "data": {"label": label},
            "position": {"x": x, "y": y},
            "draggable": True,
        }
        self.graph.nodes.append(new_node)

    def clear_graph(self):
        self.graph.nodes = initial_nodes
        self.graph.edges = initial_edges

    def on_connect(self, new_edge):
        # Iterate over the existing edges
        for i, edge in enumerate(self.graph.edges):
            # If we find an edge with the same ID as the new edge
            if edge["id"] == f"e{new_edge['source']}-{new_edge['target']}":
                # Delete the existing edge
                del self.graph.edges[i]
                break

        # Add the new edge
        self.graph.edges.append(
            {
                "id": f"e{new_edge['source']}-{new_edge['target']}",
                "source": new_edge["source"],
                "target": new_edge["target"],
                "label": random.choice(["+", "-", "*", "/"]),
                "animated": True,
            }
        )

    def on_nodes_change(self, changes):
        nodes = []
        for node in apply_changes(changes, self.graph.nodes):
            node = to_dict_if_proxy(node)
            nodes.append(node)
        self.graph.nodes = nodes



"""The Vector Store page."""

import reflex as rx

from aita_lab import styles
from aita_lab.components.upload import upload_render
from aita_lab.states.vector_store import VectorStoreState
from aita_lab.templates import ThemeState, template


def add_feature():
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.chakra.button(
                rx.flex("Add Feature +", spacing="3"),
                size="sm",
                width="100%",
                on_click=VectorStoreState.open_feature_dialog,
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(
                "Feature Information",
                size="1",
                font_family="Inter",
                padding_top="1em",
            ),
            rx.dialog.description(
                "Create a feature to store your embedding data.",
                size="2",
                mb="4",
                padding_bottom="1em",
            ),
            rx.flex(
                rx.chakra.heading("Feature View Name *", size="sm"),
                rx.input(
                    placeholder="Enter the name of the feature view (table)",
                    on_blur=VectorStoreState.set_feature_view_name,
                    required=True,
                ),
                rx.chakra.heading("Feature Name *", size="sm"),
                rx.input(
                    placeholder="Enter the feature name",
                    on_blur=VectorStoreState.set_feature_name,
                    required=True,
                ),
                rx.chakra.heading("Entities", size="sm"),
                rx.input(
                    placeholder="Enter the entities",
                    on_blur=VectorStoreState.set_feature_entities,
                    required=False,
                ),
                rx.chakra.heading("Data Source", size="sm"),
                rx.select(
                    ["files", "postgres", "elasticsearch"],
                    placeholder="Select the data source type",
                    value=VectorStoreState.data_source_type,
                    on_change=VectorStoreState.set_data_source_type,
                ),
                rx.cond(
                    VectorStoreState.data_source_type == "files",
                    upload_render(VectorStoreState.files, VectorStoreState.handle_upload),
                ),
                rx.flex(
                    rx.dialog.close(
                        rx.chakra.button(
                            "Create Feature",
                            on_click=VectorStoreState.create_vector_feature,
                        ),
                    ),
                    justify="end",
                ),
                direction="column",
                spacing="4",
            ),
        ),
    )


def setup_store():
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.button(
                rx.flex("Setup Store", rx.icon(tag="plus", width=24, height=24), spacing="3"),
                size="4",
                radius="full",
                on_click=VectorStoreState.open_store_dialog,
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(
                "Create Store",
                size="1",
                font_family="Inter",
                padding_top="1em",
            ),
            rx.dialog.description(
                "Create a Vector Store to store your embedding features.",
                size="2",
                mb="4",
                padding_bottom="1em",
            ),
            rx.flex(
                rx.chakra.heading("Project Name *", size="sm"),
                rx.input(
                    placeholder="Enter the name of the project",
                    on_blur=VectorStoreState.set_project_name,
                    required=True,
                ),
                rx.chakra.heading("Online Store Type *", size="sm"),
                rx.select(
                    ["Postgresql", "Sqlite", "Elasticsearch"],
                    placeholder="Select the online store type",
                    on_change=VectorStoreState.set_online_store_type,
                ),
                rx.chakra.heading("Online Store Configs", size="sm"),
                rx.input(
                    placeholder="Enter the online store configs",
                    on_blur=VectorStoreState.set_online_store_configs,
                ),
                rx.chakra.heading("Offline Store Type", size="sm"),
                rx.select(
                    ["file", "Sqlite", "Elasticsearch"],
                    placeholder="Select the offline store type",
                    on_change=VectorStoreState.set_offline_store_type,
                ),
                rx.chakra.heading("Offline Store Configs", size="sm"),
                rx.input(
                    placeholder="Enter the offline store configs",
                    on_blur=VectorStoreState.set_offline_store_configs,
                ),
                rx.flex(
                    rx.dialog.close(
                        rx.chakra.button(
                            "Create Store",
                            on_click=VectorStoreState.create_store,
                        ),
                    ),
                    justify="end",
                ),
                direction="column",
                spacing="4",
            ),
            open=VectorStoreState.store_dialog_open,
        ),
    )


def show_features():
    return rx.chakra.box(
        rx.chakra.heading("Embedding Features"),
        rx.chakra.text("Create a feature to store your embedding data"),
        rx.chakra.table_container(
            rx.chakra.table(
                rx.chakra.thead(
                    rx.chakra.tr(
                        rx.chakra.th("Feature View"),
                        rx.chakra.th("Entities"),
                        rx.chakra.th("Feature"),
                        rx.chakra.th("Source"),
                        rx.chakra.th("Load"),
                    )
                ),
                rx.chakra.tbody(
                    rx.foreach(
                        VectorStoreState.vector_feature_views,
                        lambda feature_view: rx.chakra.tr(
                            rx.chakra.td(feature_view.name),
                            rx.chakra.td(feature_view.entities),
                            rx.chakra.td(feature_view.feature),
                            rx.chakra.td(feature_view.source),
                            rx.chakra.td(
                                rx.chakra.button(
                                    "Load",
                                    on_click=lambda: VectorStoreState.push_source_to_feature(
                                        feature_view
                                    ),
                                )
                            ),
                        ),
                    ),
                    width="100%",
                ),
                color_scheme="white",
                variant="striped",
                width="100%",
            ),
            add_feature(),
            width="100%",
        ),
        width="100%",
        padding="1em",
    )


def show_store():
    return rx.chakra.box(
        rx.tabs.root(
            rx.tabs.list(
                rx.foreach(
                    VectorStoreState.project_names,
                    lambda project_name: rx.tabs.trigger(
                        project_name,
                        value=project_name,
                        cursor="pointer",
                    ),
                ),
            ),
            rx.tabs.content(
                show_features(),
                value=VectorStoreState.project_name,
                width="100%",
            ),
            value=VectorStoreState.project_name,
            on_change=VectorStoreState.set_project,
            orientation="horizontal",
            height="100vh",
            width="100%",
        ),
        width="100%",
        border=styles.border,
        border_radius=styles.border_radius,
    )


@template(route="/vector_store", title="Vector Store", on_load=[VectorStoreState.on_load()])
def vector_store() -> rx.Component:
    """The Vector Store page.

    Returns:
        The UI for the Vector Store page.
    """
    return rx.vstack(
        rx.chakra.heading("Vector Store"),
        rx.chakra.text("Create a vector store to store your embedding features"),
        setup_store(),
        show_store(),
        width="100%",
    )



from typing import Dict, Optional

from aita.agent.graph import GraphAgent
from aita.tool.python_tool import PythonTool


class PythonAgent(GraphAgent):
    type: str = "python"
    description: str = "A Python agent that can use Python tools to run python scripts."

    def __init__(
        self,
        model: str,
        model_parameters: Optional[Dict] = None,
    ):
        super().__init__([PythonTool()], model, model_parameters)

    def add_script_context(self, script_context):
        for tool in self.tools:
            tool.update_script_context(script_context=script_context)
        return self



"""Welcome to Reflex!."""

# Import all the pages.
import reflex as rx

from aita_lab.pages import *


class State(rx.State):
    """Define empty state to allow access to rx.State.router."""


# Create the app.
app = rx.App()



from typing import List, Optional

import logging

import reflex as rx
from sqlmodel import select

from aita.datasource.factory import DataSourceFactory, DataSourceProvider
from aita_lab.states.catalog import CatalogState


class DataSource(rx.Model, table=True):
    """The SqlDataSource model."""

    name: str
    datasource_type: str
    connection_url: str


class DataSourceState(rx.State):
    id: str
    name: str
    datasource_names: list[str] = []
    datasource_type: str
    datasources: list[DataSource] = []
    num_datasources: int
    sort_value: str
    is_open: bool = False
    allow_crawl_catalog: bool = True

    # DataSource related attributes
    connection_url: str
    open_alert: bool = False

    def toggle_dialog(self):
        self.is_open = not self.is_open

    def change_crawl_catalog(self, value: bool):
        self.allow_crawl_catalog = value

    @rx.var
    def datasource_attributes(self) -> List[str]:
        if self.datasource_type:
            model_fields = DataSourceProvider[self.datasource_type].value.__fields__
            return [key for key in model_fields.keys() if key != "type" and key != "name"]
        else:
            return []

    @rx.var
    def missing_attributes(self) -> bool:
        # required by all datasources
        if not self.datasource_type or not self.name:
            return True

        # required by specific datasource
        model_fields = DataSourceProvider[self.datasource_type].value.__fields__
        if not all(
            getattr(self, key) for key in model_fields.keys() if key != "type" and key != "name"
        ):
            return True
        return False

    def set_datasource_attributes(self, attribute: str, value: str):
        setattr(self, attribute, value)

    def get_datasource_attributes(self) -> Optional[dict]:
        if self.datasource_type:
            model_fields = DataSourceProvider[self.datasource_type].value.__fields__
            return {key: getattr(self, key) for key in model_fields.keys() if key != "type"}
        else:
            return None

    def load_entries(self):
        with rx.session() as session:
            self.datasources = session.exec(select(DataSource)).all()
            self.num_datasources = len(self.datasources)

            if self.sort_value:
                self.datasources = sorted(
                    self.datasources,
                    key=lambda datasource: getattr(datasource, self.sort_value).lower(),
                )
            self.datasource_names = [datasource.name for datasource in self.datasources]

    def sort_values(self, sort_value: str):
        self.sort_value = sort_value
        self.load_entries()

    def set_datasource(self, datasource: DataSource):
        self.id = datasource["id"]
        self.name = datasource["name"]
        self.datasource_type = datasource["datasource_type"]
        self.connection_url = datasource["connection_url"]

    def connect_and_add_datasource(self):
        if self.missing_attributes:
            return
        config = self.get_datasource_attributes()
        try:
            ds = DataSourceFactory.create_datasource(self.datasource_type, **config)
            ds.connect()
            logging.info(f"Connected to {self.datasource_type}")
        except Exception as e:
            logging.error(f"Failed to connect to {self.datasource_type}: {e}")
            return
        if self.allow_crawl_catalog:
            CatalogState.crawl_data_catalog(self.name, ds)
        with rx.session() as session:
            session.add(
                DataSource(
                    name=self.name,
                    datasource_type=self.datasource_type,
                    connection_url=self.connection_url,
                )
            )
            session.commit()
        self.load_entries()
        self.toggle_dialog()

    def update_datasource(self):
        with rx.session() as session:
            datasource = session.exec(select(DataSource).where(DataSource.id == self.id)).first()
            datasource.name = self.name
            datasource.connection_url = self.connection_url
            session.add(datasource)
            session.commit()
        self.load_entries()

    def delete_datasource(self, datasource_id: str):
        with rx.session() as session:
            datasource = session.exec(
                select(DataSource).where(DataSource.id == datasource_id)
            ).first()
            session.delete(datasource)
            session.commit()
        self.load_entries()

    @staticmethod
    def get_datasource_by_name(datasource_name: str):
        with rx.session() as session:
            datasource = session.exec(
                select(DataSource).where(DataSource.name == datasource_name)
            ).first()
            return datasource

    @staticmethod
    def connect(datasource_name: str) -> Optional[DataSource]:
        datasource = DataSourceState.get_datasource_by_name(datasource_name)
        if not datasource:
            return
        try:
            source = DataSourceFactory.create_datasource(
                datasource.datasource_type, connection_url=datasource.connection_url
            )
            source.connect()
            logging.info(f"Connected to {datasource.datasource_type}")
            return source
        except Exception as e:
            logging.error(f"Failed to connect to {datasource.datasource_type}: {e}")

    def on_load(self):
        self.load_entries()



from typing import Optional

from sqlalchemy import Engine
from sqlalchemy.engine import URL, create_engine

from aita.datasource.catalog import Catalog
from aita.datasource.sql import SqlDataSource


class MySqlDataSource(SqlDataSource):

    def __init__(self, connection_url: Optional[str] = None, **kwargs):
        if not connection_url:
            user = kwargs.get("user")
            password = kwargs.get("password")
            host = kwargs.get("host")
            port = kwargs.get("port")
            database = kwargs.get("database")
            connection_url = self.build_connection_url(
                user=user,
                password=password,
                host=host,
                port=port,
                database=database,
                **kwargs,
            )
        super().__init__(connection_url, "mysql")

    def build_connection_url(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        **kwargs,
    ) -> str:
        return URL(
            drivername="mysql+pymysql",
            username=user,
            password=password,
            host=host,
            port=port,
            database=database,
            query=kwargs,
        ).render_as_string()

    def connect(self) -> Engine:
        return create_engine(self.connection_url)

    def get_metadata(self, **kwargs) -> Catalog:
        raise NotImplementedError



import types

import asyncio


async def helper9():
    await hello_world_message()


async def hello_world_message() -> str:
    await asyncio.sleep(1)
    return "Hello, world!"


async def main() -> None:
    message = await helper9()
    print(message)


asyncio.run(main())



from typing import Optional

import logging

import reflex as rx
from sqlmodel import select


class CatalogTable(rx.Model, table=True):
    """The Catalog Table Model."""

    datasource_name: str
    catalog_name: str
    schema: str
    table: str


class CatalogState(rx.State):
    catalogs: list[CatalogTable] = []

    is_open: bool = False
    sort_value: str = ""

    def toggle_dialog(self):
        self.is_open = not self.is_open

    def add_catalog(self, catalog: CatalogTable):
        self.commit(*catalog)
        self.load_entries()

    def delete_catalog(self, catalog: CatalogTable):
        self.catalogs.remove(catalog)

    def sort_values(self, value: str):
        self.sort_value = value
        self.catalogs = sorted(self.catalogs, key=lambda x: getattr(x, value))

    def load_entries(self):
        with rx.session() as session:
            self.catalogs = session.exec(select(CatalogTable)).all()

            if self.sort_value:
                self.catalogs = sorted(
                    self.catalogs,
                    key=lambda catalog: getattr(catalog, self.sort_value).lower(),
                )

    @staticmethod
    def commit(catalog: str, schema: str, table: str, datasource_name: Optional[str] = None):
        with rx.session() as session:
            session.add(
                CatalogTable(
                    catalog_name=catalog,
                    schema=schema,
                    table=table,
                    datasource_name=datasource_name,
                )
            )
            session.commit()

    @staticmethod
    def crawl_data_catalog(datasource_name, datasource):
        catalog = datasource.get_metadata()
        print(catalog)
        for schema in catalog.catalog_db_schemas:
            for table in schema.db_schema_tables:
                try:
                    CatalogState.commit(
                        catalog=catalog.catalog_name,
                        schema=schema.db_schema_name,
                        table=table.table_name,
                        datasource_name=datasource_name,
                    )
                except Exception as e:
                    logging.error(f"Failed to crawl data catalog: {e}")

    def on_load(self):
        self.load_entries()
